#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Functionality common to TableViews and FormViews"""

from PyQt4 import QtGui
from PyQt4 import QtCore

from camelot.view.model_thread import post
from camelot.view.model_thread import model_function


class AbstractView(QtGui.QWidget):
    """A string used to format the title of the view ::
    title_format = 'Movie rental overview'

    .. attribute:: header_widget

    The widget class to be used as a header in the table view::

    header_widget = None
    """

    title_format = ''
    header_widget = None

    title_changed_signal = QtCore.pyqtSignal(QtCore.QString)
    icon_changed_signal = QtCore.pyqtSignal(QtGui.QIcon)

    @QtCore.pyqtSlot()
    def refresh(self):
        """Refresh the data in the current view"""
        pass

    @QtCore.pyqtSlot(object)
    def change_title(self, new_title):
        """Will emit the title_changed_signal"""
        #import sip
        #if not sip.isdeleted(self):
        self.title_changed_signal.emit( unicode(new_title) )
        
    @QtCore.pyqtSlot(object)
    def change_icon(self, new_icon):
        self.icon_changed_signal.emit(new_icon)

    @model_function
    def to_html(self):
        pass

    @model_function
    def export_to_word(self):
        from camelot.view.export.word import open_html_in_word
        html = self.to_html()
        open_html_in_word(html)

    @model_function
    def export_to_excel(self):
        from camelot.view.export.excel import open_data_with_excel
        title = self.getTitle()
        columns = self.getColumns()
        data = [d for d in self.getData()]
        open_data_with_excel(title, columns, data)

    @model_function
    def export_to_mail(self):
        from camelot.view.export.outlook import open_html_in_outlook
        html = self.to_html()
        open_html_in_outlook(html)

class TabView(AbstractView):
    """Class to combine multiple views in Tabs and let them behave as one view.
    This class can be used when defining custom create_table_view methods on an
    ObjectAdmin class to group multiple table views together in one view."""

    def __init__(self, parent, views=[], admin=None):
        """:param views: a list of the views to combine"""
        AbstractView.__init__(self, parent)
        layout = QtGui.QVBoxLayout()
        if self.header_widget:
            self.header = self.header_widget(self, admin)
        else:
            self.header = None
        layout.addWidget(self.header)
        self._tab_widget = QtGui.QTabWidget(self)
        self._tab_widget.setObjectName( 'tab_widget' )
        layout.addWidget(self._tab_widget)
        self.setLayout(layout)

        def get_views_and_titles():
            return [(view, view.get_title()) for view in views]

        post(get_views_and_titles, self.set_views_and_titles)
        post(lambda:self.title_format, self.change_title)

    @QtCore.pyqtSlot()
    def refresh(self):
        """Refresh the data in the current view"""
        for i in range(self._tab_widget.count()):
            view = self._tab_widget.widget(i)
            view.refresh()

    def set_views_and_titles(self, views_and_titles):
        for view, title in views_and_titles:
            self._tab_widget.addTab(view, title)

    def export_to_excel(self):
        return self._tab_widget.currentWidget().export_to_excel()

    def export_to_word(self):
        return self._tab_widget.currentWidget().export_to_word()

    def export_to_mail(self):
        return self._tab_widget.currentWidget().export_to_mail()

    def to_html(self):
        return self._tab_widget.currentWidget().to_html()





import os

from PyQt4 import QtGui

from camelot.admin.action import Action
from camelot.admin.action import list_action
from camelot.core.utils import ugettext_lazy as _
from camelot.test import ModelThreadTestCase
from camelot.test.action import MockModelContext

from test_view import static_images_path

class ActionWidgetsCase( ModelThreadTestCase ):
    """Test widgets related to actions.
    """

    images_path = static_images_path

    def setUp(self):
        from camelot.admin.action import ApplicationActionGuiContext, State
        from camelot.admin.application_admin import ApplicationAdmin
        from camelot_example.importer import ImportCovers
        ModelThreadTestCase.setUp(self)
        self.app_admin = ApplicationAdmin()
        self.action = ImportCovers()
        self.application_gui_context = ApplicationActionGuiContext()
        self.parent = QtGui.QWidget()
        enabled = State()
        disabled = State()
        disabled.enabled = False
        notification = State()
        notification.notification = True
        self.states = [ ( 'enabled', enabled),
                        ( 'disabled', disabled),
                        ( 'notification', notification) ]
        
    def grab_widget_states( self, widget, suffix ):
        for state_name, state in self.states:
            widget.set_state( state )
            self.grab_widget( widget, suffix='%s_%s'%( suffix,
                                                       state_name ) )
        
    def test_action_label( self ):
        from camelot.view.controls.action_widget import ActionLabel
        widget = ActionLabel( self.action,
                              self.application_gui_context,
                              self.parent )
        self.grab_widget_states( widget, 'application' )

    def test_action_push_botton( self ):
        from camelot.view.controls.action_widget import ActionPushButton
        widget = ActionPushButton( self.action,
                                   self.application_gui_context,
                                   self.parent )
        self.grab_widget_states( widget, 'application' )
        
class ActionStepsCase( ModelThreadTestCase ):
    """Test the various steps that can be executed during an
    action.
    """

    images_path = static_images_path
    
    def setUp(self):
        from camelot_example.model import Movie
        from camelot.admin.application_admin import ApplicationAdmin
        ModelThreadTestCase.setUp(self)
        self.app_admin = ApplicationAdmin()
        self.context = MockModelContext()
        self.context.obj = Movie.query.first()

# begin test application action
    def test_example_application_action( self ):
        from camelot_example.importer import ImportCovers
        from camelot_example.model import Movie
        # count the number of movies before the import
        movies = Movie.query.count()
        # create an import action
        action = ImportCovers()
        generator = action.model_run( None )
        select_file = generator.next()
        self.assertFalse( select_file.single )
        # pretend the user selected a file
        generator.send( [os.path.join( os.path.dirname(__file__), '..', 'camelot_example', 'media', 'covers', 'circus.jpg') ] )
        # continue the action till the end
        list( generator )
        # a movie should be inserted
        self.assertEqual( movies + 1, Movie.query.count() )
# end test application action

    def test_change_object( self ):
        from camelot.bin.meta import NewProjectOptions
        from camelot.view.action_steps.change_object import ChangeObjectDialog
        admin = NewProjectOptions.Admin( self.app_admin, NewProjectOptions )
        options = NewProjectOptions()
        options.name = 'Videostore'
        options.module = 'videostore'
        options.domain = 'example.com'
        dialog = ChangeObjectDialog( options, admin )
        self.grab_widget( dialog )
        
    def test_select_file( self ):
        from camelot.view.action_steps import SelectFile
        select_file = SelectFile( 'Image Files (*.png *.jpg);;All Files (*)' )
        dialog = select_file.render()
        self.grab_widget( dialog )
        
    def test_print_preview( self ):
        
        # begin webkit print
        class WebkitPrint( Action ):
            
            def model_run( self, model_context ):
                from PyQt4.QtWebKit import QWebView
                from camelot.view.action_steps import PrintPreview
                
                movie = model_context.get_object()
                
                document = QWebView()
                document.setHtml( '<h2>%s</h2>' % movie.title )
                
                yield PrintPreview( document )
        # end webkit print
                
        action = WebkitPrint()
        steps = list( action.model_run( self.context ) )
        dialog = steps[0].render()
        dialog.show()
        self.grab_widget( dialog )
        
    def test_print_html( self ):
        
        # begin html print
        class MovieSummary( Action ):
            
            verbose_name = _('Summary')
            
            def model_run(self, model_context):
                from camelot.view.action_steps import PrintHtml
                movie = model_context.get_object()
                yield PrintHtml( "<h1>This will become the movie report of %s!</h1>" % movie.title )
        # end html print
 
        action = MovieSummary()
        steps = list( action.model_run( self.context ) )
        dialog = steps[0].render()
        dialog.show()
        self.grab_widget( dialog )

class ListActionsCase( ModelThreadTestCase ):
    """Test the standard list actions.
    """

    images_path = static_images_path

    def setUp( self ):
        from camelot_example.model import Movie
        from camelot.admin.application_admin import ApplicationAdmin
        ModelThreadTestCase.setUp(self)
        self.app_admin = ApplicationAdmin()
        self.context = MockModelContext()
        self.context.obj = Movie.query.first()
        self.context.admin = self.app_admin.get_related_admin( Movie )
        
    def test_change_row_actions( self ):
        from camelot.test.action import MockListActionGuiContext
        
        gui_context = MockListActionGuiContext()
        get_state = lambda action:action.get_state( gui_context.create_model_context() )
        to_first = list_action.ToFirstRow()
        to_previous = list_action.ToPreviousRow()
        to_next = list_action.ToNextRow()
        to_last = list_action.ToLastRow()
        
        to_last.gui_run( gui_context )
        self.assertFalse( get_state( to_last ).enabled )
        self.assertFalse( get_state( to_next ).enabled )
        to_previous.gui_run( gui_context )
        self.assertTrue( get_state( to_last ).enabled )
        self.assertTrue( get_state( to_next ).enabled )
        to_first.gui_run( gui_context )
        self.assertFalse( get_state( to_first ).enabled )
        self.assertFalse( get_state( to_previous ).enabled )
        to_next.gui_run( gui_context )
        self.assertTrue( get_state( to_first ).enabled )
        self.assertTrue( get_state( to_previous ).enabled )
        
    def test_print_preview( self ):
        print_preview = list_action.PrintPreview()
        for step in print_preview.model_run( self.context ):
            dialog = step.render()
            dialog.show()
            self.grab_widget( dialog )



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Camelot specific subclasses of Exception
"""

from camelot.core.utils import ugettext_lazy as _

class UserException(Exception):
    """
    Raise this exception to inform the user he did something wrong, without
    showing a stack trace or other internals.  Raising this exception won't
    log stack traces either, as the occurance of this exception is considered
    a non-event for the developer::
    
        from camelot.core.exception import UserException
        from camelot.core.utils import ugettext
        
        if not dvd.empty:
            raise UserException( ugettext('Could not burn movie to non empty DVD'),
                                 resolution = ugettext('Insert an empty DVD and retry') )

    Will popup a gentle dialog for the user :

    .. image:: /_static/controls/user_exception.png

    """
    
    def __init__(self, text, title=_('Could not proceed'), icon=None, resolution=None, detail=None):
        """
        :param title: the title of the dialog box that informs the user
        :param text: the top text in the dialog
        :param resolution: what the user should do to solve the issue
        :param detail: a detailed description of what went wrong
        """
        super(UserException, self).__init__(text)
        self.title = title
        self.text = text
        self.icon = icon
        self.resolution = resolution
        self.detail = detail

class GuiException(Exception):
    """
    This exception is raised by the Action mechanism when the action requested
    something from the GUI but an unexpected event occured.  The action can
    choose to ignore it or handle it.
    """
    pass

class CancelRequest(Exception):
    """
    This exception is raised by the GUI when the user wants to cancel an action,
    this exception is then past to the *model thread*
    """
    pass



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""left navigation pane"""

import logging
logger = logging.getLogger('camelot.view.controls.navpane2')

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import Qt
from PyQt4.QtGui import QMenu
from PyQt4.QtGui import QFrame
from PyQt4.QtGui import QWidget
from PyQt4.QtGui import QToolBox
from PyQt4.QtGui import QDockWidget
from PyQt4.QtGui import QVBoxLayout

from camelot.admin.action.application_action import ApplicationActionGuiContext
from camelot.core.utils import variant_to_pyobject
from camelot.view.model_thread import post
from camelot.view.controls.modeltree import ModelItem
from camelot.view.controls.modeltree import ModelTree

class PaneSection(QWidget):

    def __init__(self, parent, section, workspace):
        super(PaneSection, self).__init__(parent)
        self._items = []
        self._workspace = workspace
        self._section = section
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        section_tree = ModelTree(parent=self)
        # i hate the sunken frame style
        section_tree.setFrameShape(QFrame.NoFrame)
        section_tree.setFrameShadow(QFrame.Plain)
        section_tree.contextmenu = QMenu(self)
        section_tree.setContextMenuPolicy(Qt.CustomContextMenu)
        section_tree.customContextMenuRequested.connect(self.create_context_menu)
        section_tree.setObjectName( 'SectionTree' )
        section_tree.itemClicked.connect( self._item_clicked )
        section_tree.setWordWrap( False )
        layout.addWidget( section_tree )
        self.setLayout(layout)
        post( section.get_items, self.set_items )

    @QtCore.pyqtSlot(object)
    def set_items(self, items):
        logger.debug('setting items for current navpane section')
        section_tree = self.findChild(QtGui.QWidget, 'SectionTree')
        self._items = items
        if section_tree:
            section_tree.clear()
            section_tree.clear_model_items()
    
            if not items: return
    
            for item in items:
                label = item.get_verbose_name()
                icon = item.get_icon()
                model_item = ModelItem( section_tree, 
                                        [unicode(label)],
                                        item )
                if icon:
                    model_item.set_icon(icon.getQIcon())
                section_tree.modelitems.append( model_item )
            section_tree.resizeColumnToContents( 0 )

    def create_context_menu(self, point):
        logger.debug('creating context menu')
        section_tree = self.findChild(QtGui.QWidget, 'SectionTree')
        if section_tree:
            item = section_tree.itemAt(point)
            if item:
                section_tree.contextmenu.clear()
                for mode in item.section_item.get_modes():
                    action = mode.render( self )
                    action.triggered.connect( self._action_triggered )
                    section_tree.contextmenu.addAction( action )
                section_tree.setCurrentItem(item)
                section_tree.contextmenu.popup(section_tree.mapToGlobal(point))

    @QtCore.pyqtSlot(bool)
    def _action_triggered( self, _checked ):
        action = self.sender()
        mode_name = variant_to_pyobject( action.data() )
        self._run_current_action( mode_name )
        
    @QtCore.pyqtSlot(QtGui.QTreeWidgetItem, int)
    def _item_clicked(self, _item, _column):
        self._run_current_action()

    def _run_current_action( self, mode_name=None ):
        section_tree = self.findChild(QtGui.QWidget, 'SectionTree')
        if section_tree:
            item = section_tree.currentItem()
            index = section_tree.indexFromItem(item)
            section_item = self._items[index.row()]
            gui_context = ApplicationActionGuiContext()
            gui_context.mode_name = mode_name
            gui_context.workspace = self._workspace
            gui_context.admin = self._section.admin
            section_item.get_action().gui_run( gui_context )
                        
class NavigationPane(QDockWidget):

    def __init__(self, app_admin, workspace, parent):
        super(NavigationPane, self).__init__(parent)

        self._workspace = workspace
        self.app_admin = app_admin
        
        tb = QToolBox()
        tb.setFrameShape(QFrame.NoFrame)
        tb.layout().setContentsMargins(0,0,0,0)
        tb.layout().setSpacing(1)
        tb.setObjectName('toolbox')
        tb.setMouseTracking(True)
        
        # hack for removing the dock title bar
        self.setTitleBarWidget(QWidget())
        self.setWidget(tb)
        self.setFeatures(QDockWidget.NoDockWidgetFeatures)

        self.app_admin.sections_changed_signal.connect(self.update_sections)
        self.update_sections()

    def wheelEvent(self, wheel_event):
        steps = -1 * wheel_event.delta() / (8 * 15)
        toolbox = self.findChild(QtGui.QWidget, 'toolbox')
        if steps and toolbox:
            current_index = toolbox.currentIndex()
            toolbox.setCurrentIndex( max( 0, min( current_index + steps, toolbox.count() ) ) )
        
    @QtCore.pyqtSlot()
    def update_sections(self):
        post(self.app_admin.get_sections, self.set_sections)

    def get_sections(self):
        return self._sections

    @QtCore.pyqtSlot(object)
    def set_sections(self, sections):
        logger.debug('setting navpane sections')
        if not sections:
            self.setMaximumWidth(0)
            return
        toolbox = self.findChild(QtGui.QWidget, 'toolbox')
        animation = QtCore.QPropertyAnimation(toolbox, 'minimumWidth', self)
        animation.setDuration( 500 )
        animation.setStartValue( 0 )
        animation.setEndValue( 220 )

        if self._workspace:
            animation.finished.connect(self._workspace._background_widget.makeInteractive)

        animation.start()

        # performs QToolBox clean up
        # QToolbox won't delete items we have to do it explicitly
        count = toolbox.count()
        while count:
            item = toolbox.widget(count-1)
            toolbox.removeItem(count-1)
            item.deleteLater()
            count -= 1
            
        for section in sections:
            # TODO: old navpane used translation here
            name = unicode( section.get_verbose_name() )
            icon = section.get_icon().getQIcon()
            pwdg = PaneSection(toolbox, section, self._workspace)
            toolbox.addItem(pwdg, icon, name)

        toolbox.setCurrentIndex(0)
        # WARNING: hardcoded width
        #self._toolbox.setMinimumWidth(220)




#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
"""Most users have the need to do some basic task tracking across various
parts of the data model.

These classes provide basic task tracking with configurable statuses, 
categories and roles.  They are presented to the user as "Todo's"
"""

from elixir import Entity, using_options, Field, ManyToMany, OneToMany, ManyToOne, entities, ColumnProperty
import sqlalchemy.types
from sqlalchemy import sql
from sqlalchemy.orm import backref

from camelot.core.utils import ugettext_lazy as _
from camelot.model import metadata
from camelot.model.authentication import getCurrentAuthentication
from camelot.model.type_and_status import type_3_status, create_type_3_status_mixin, get_status_type_class, get_status_class
from camelot.admin.entity_admin import EntityAdmin
from camelot.core.document import documented_entity
from camelot.view import forms
from camelot.view.filters import ComboBoxFilter
from camelot.view.controls import delegates
import camelot.types

import datetime

__metadata__ = metadata

def task_has_id(task):
    return task.id

task_status_type_features = [(None, None, '', ()),
                             (1, 'hidden', 'Zichtbaar in lijst', ((0, 'False'), (1, 'True'))),
                             (2, 'default', '', ((0, 'False'), (1, 'True')))]
task_status_type_features_enumeration = [(f[0], f[1]) for f in task_status_type_features]
task_status_type_feature_descriptions = dict((f[1], f[2]) for f in task_status_type_features)
task_status_type_feature_values = dict((f[1], f[3]) for f in task_status_type_features)

class TaskType( Entity ):
    using_options(tablename='task_type', order_by=['rank', 'description'])
    description = Field( sqlalchemy.types.Unicode(48), required=True, index=True )
    rank = Field(sqlalchemy.types.Integer(), default=1)
    
    def __unicode__(self):
        return self.description or ''
    
    class Admin( EntityAdmin ):
        verbose_name = _('Task Type')
        list_display = ['description', 'rank']

class TaskRoleType( Entity ):
    using_options(tablename='task_role_type', order_by=['rank', 'description'])
    description = Field( sqlalchemy.types.Unicode(48), required=True, index=True )
    rank = Field(sqlalchemy.types.Integer(), default=1)
    
    def __unicode__(self):
        return self.description or ''
    
    class Admin( EntityAdmin ):
        verbose_name = _('Task Role Type')
        list_display = ['description', 'rank']
        

class TaskRole( Entity ):
    using_options(tablename='task_role')
    task = ManyToOne('Task', required = True, ondelete = 'cascade', onupdate = 'cascade')
    party = ManyToOne('Party', required=True, ondelete='restrict', onupdate='cascade')
    described_by = ManyToOne('TaskRoleType', required = False, ondelete = 'restrict', onupdate = 'cascade')
    rank = Field(sqlalchemy.types.Integer(), required=True, default=1)
    comment = Field( sqlalchemy.types.Unicode( 256 ) )

    class Admin(EntityAdmin):
        verbose_name = _('Role within task')
        list_display = ['party', 'described_by', 'comment', 'rank']
        field_attributes = {'described_by':{'name':_('Type'), 'delegate':delegates.ManyToOneChoicesDelegate},
                                 'rank':{'choices':[(i,str(i)) for i in range(1,5)]},
                                 }

class TaskNote( Entity ):
    using_options(tablename='task_note', order_by=['-created_at'] )
    of = ManyToOne('Task', required=True, onupdate='cascade', ondelete='cascade')
    created_at = Field( sqlalchemy.types.Date, required=True, default=datetime.date.today )
    created_by = ManyToOne('AuthenticationMechanism', required=True )
    note = Field( camelot.types.RichText() )
  
    class Admin( EntityAdmin ):
        verbose_name = _('Note')
        list_display = ['created_at', 'created_by']
        form_display = list_display + ['note']

class TaskDocumentType( Entity ):
    using_options(tablename='task_document_type', order_by=['rank', 'description'])
    description = Field( sqlalchemy.types.Unicode(48), required=True, index=True )
    rank = Field(sqlalchemy.types.Integer(), default=1)
        
    def __unicode__(self):
        return self.description or ''
    
    class Admin( EntityAdmin ):
        verbose_name = _('Document Type')
        list_display = ['description', 'rank']
  
class TaskDocument( Entity ):
    using_options(tablename='task_document')
    of = ManyToOne('Task', required=True, onupdate='cascade', ondelete='cascade')
    created_at = Field( sqlalchemy.types.Date(), default = datetime.date.today, required = True, index = True )
    created_by = ManyToOne('AuthenticationMechanism', required=True)
    type = ManyToOne('TaskDocumentType', required = False, ondelete = 'restrict', onupdate = 'cascade')
    document = Field( camelot.types.File(), required=True )
    description = Field( sqlalchemy.types.Unicode(200) )
    summary = Field( camelot.types.RichText() )
        
    class Admin(EntityAdmin):
        verbose_name = _('Document')
        list_display = ['document', 'description', 'type']
        form_display = list_display + ['created_at', 'created_by', 'summary']
        field_attributes = {'type':{'delegate':delegates.ManyToOneChoicesDelegate},
                            'created_by':{'default':getCurrentAuthentication}}


class Task( Entity, create_type_3_status_mixin('status') ):
    using_options(tablename='task', order_by=['-creation_date', 'id'] )
    creation_date    = Field( sqlalchemy.types.Date, required=True, default=datetime.date.today )
    due_date         = Field( sqlalchemy.types.Date, required=False, default=None )
    description      = Field( sqlalchemy.types.Unicode(255), required=True )
    status           = OneToMany( type_3_status( 'Task', metadata, entities ), cascade='all, delete, delete-orphan' )
    notes            = OneToMany( 'TaskNote', cascade='all, delete, delete-orphan' )
    documents        = OneToMany( 'TaskDocument', cascade='all, delete, delete-orphan' )
    roles            = OneToMany( 'TaskRole', cascade='all, delete, delete-orphan' )
    described_by     = ManyToOne( 'TaskType', required = False, ondelete = 'restrict', onupdate = 'cascade')
    categories       = ManyToMany( 'PartyCategory',
                                   tablename='party_category_task', 
                                   remote_colname='party_category_id',
                                   local_colname='task_id')
                             
    Admin = None
    
    @ColumnProperty
    def number_of_documents( self ):
        return sql.select( [sql.func.count( TaskDocument.id ) ],
                            whereclause = TaskDocument.of_id == self.id )

    @ColumnProperty
    def current_status_sql( self ):
        status_class = get_status_class('Task')
        status_type_class = get_status_type_class('Task')
        return sql.select( [status_type_class.code],
                           whereclause = sql.and_( status_class.status_for_id == self.id,
                                                   status_class.status_from_date <= sql.functions.current_date(),
                                                   status_class.status_thru_date >= sql.functions.current_date() ),
                           from_obj = [status_type_class.table.join( status_class.table )] ).limit(1)

    @ColumnProperty
    def hidden(self):
        status_class = get_status_class('Task')
        status_type_class = get_status_type_class('Task')
        query = sql.select( [TaskStatusTypeFeature.value],
                            whereclause = sql.and_( status_class.status_for_id == self.id,
                                                    status_class.status_from_date <= sql.functions.current_date(),
                                                    status_class.status_thru_date >= sql.functions.current_date(),
                                                    TaskStatusTypeFeature.described_by == 'hidden' ),
                            from_obj = [status_type_class.table.join( status_class.table ).join( TaskStatusTypeFeature.table )] ).limit(1)
        return sql.func.coalesce( query.as_scalar(), 0 )
        
    @classmethod
    def role_query(cls, columns, role_type_rank ):
        from camelot.model.authentication import Party
        return sql.select( [Party.full_name],
                            sql.and_( Party.id == TaskRole.party_id,
                                      TaskRole.task_id == columns.id,
                                      TaskRoleType.id == TaskRole.described_by_id,
                                      TaskRoleType.rank == role_type_rank ) ).limit(1)
    
    @ColumnProperty
    def role_1(self):
        return Task.role_query( self, 1 )
    
    @ColumnProperty
    def role_2(self):
        return Task.role_query( self, 2 )
    
    @property
    def documents_icon(self):
        if (self.number_of_documents > 0) or len(self.documents):
            return 'document'

    def __unicode__( self ):
        return self.description or ''
    
    def _get_first_note(self):
        if self.notes:
            return self.notes[0].note

    def _set_first_note(self, note):
        if note and self.id:
            if self.notes:
                self.notes[0].note = note
            else:
                self.notes.append( TaskNote( note=note, created_by=getCurrentAuthentication() ) )
 
    note = property( _get_first_note, _set_first_note )
                                
            
TaskStatusType = get_status_type_class( 'Task' )
Task = documented_entity()( Task )

class TaskStatusTypeAdmin(TaskStatusType.Admin):
    def get_delete_message(self, obj):
        return _('All tasks that are related to this status will be effected by this removal.\n\n' +
                 'Are you sure you want to remove this status?')
                  
TaskStatusType.Admin = TaskStatusTypeAdmin
TaskStatusType.Admin.delete_mode = 'on_confirm'

class TaskAdmin( EntityAdmin ):
    verbose_name = _('Task')
    verbose_name_plural = _('Tasks')
    list_display = ['creation_date', 'due_date', 'description', 'described_by', 'current_status_sql', 'role_1', 'role_2', 'documents_icon']
    list_filter  = [ComboBoxFilter('described_by.description'), 
                    ComboBoxFilter('current_status_sql'), 
                    ComboBoxFilter('categories.name'), 'hidden']
    form_state = 'maximized'
    form_display = forms.TabForm( [ ( _('Task'), ['description', 'described_by', 'current_status', 
                                                  'creation_date', 'due_date',  'note',]),
                                    ( _('Category'), ['categories'] ),
                                    ( _('Roles'), ['roles'] ),
                                    ( _('Documents'), ['documents'] ),
                                    ( _('Status'), ['status'] ) ] )
    field_attributes = {'note':{'delegate':delegates.RichTextDelegate,
                                'editable':lambda self:self.id},
                        'described_by':{'delegate':delegates.ManyToOneChoicesDelegate, 'name':_('Type')},
                        'role_1':{'editable':False},
                        'role_2':{'editable':False},
                        'description':{'minimal_column_width':50},
                        'current_status_sql':{'name':'Current status',
                                              'editable':False}}
    
    def get_field_attributes(self, field_name):
        field_attributes = EntityAdmin.get_field_attributes(self, field_name)
        if field_name in ['role_1', 'role_2']:
            name_query = sql.select( [TaskRoleType.description], TaskRoleType.rank == int(field_name[-1]) ).limit(1)
            name = TaskRoleType.query.session.scalar( name_query )
            field_attributes['name'] = name or field_attributes['name']
        return field_attributes

    def flush(self, obj):
        """Set the status of the agreement to draft if it has no status yet"""
        #if not len(obj.status):
            #obj.status.append(self.get_field_attributes('status')['target'](status_from_date=datetime.date.today(),
                                                                            #status_thru_date=end_of_times(),
                                                                            #classified_by=TaskStatusType.query.order_by(TaskStatusType.).first()))
        EntityAdmin.flush(self, obj)
  
Task.Admin = TaskAdmin                      

class TaskStatusTypeFeature(Entity):
    using_options(tablename='task_status_type_feature')
    task_status_type_id = ManyToOne( TaskStatusType, required = True, ondelete = 'cascade', onupdate = 'cascade',
                                     backref=backref('available_with', cascade="all, delete-orphan"))
    described_by = Field(camelot.types.Enumeration(task_status_type_features_enumeration), required = True, default = "hidden")
    comment = Field( sqlalchemy.types.Unicode( 256 ) )
    value = Field( sqlalchemy.types.Integer() )
    
    class Admin(EntityAdmin):
        verbose_name = _('TaskStatusTypeFeature')
        list_display = ['described_by', 'comment', 'value']
        form_display = list_display + ['comment']
        field_attributes = {'described_by': {'name': _('Feature'), 
                                             'tooltip': lambda o:task_status_type_feature_descriptions.get(o.described_by, None)},
                            'value': {'choices': lambda f:task_status_type_feature_values[f.described_by]},
                            }
                            
TaskStatusType.Admin.form_display.append('available_with')



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.core.utils import ugettext as _, variant_to_pyobject
from camelot.view.proxy import ValueLoading

class BoolDelegate(CustomDelegate):
    """Custom delegate for boolean values"""
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.BoolEditor
  
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        checked = index.model().data(index, Qt.EditRole).toBool()
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        check_option = QtGui.QStyleOptionButton()
        
        rect = QtCore.QRect(option.rect.left(),
                            option.rect.top(),
                            option.rect.width(),
                            option.rect.height())
        
        check_option.rect = rect
        check_option.palette = option.palette
        if (option.state & QtGui.QStyle.State_Selected):
            painter.fillRect(option.rect, option.palette.highlight())
        elif not self.editable:
            painter.fillRect(option.rect, option.palette.window())
        else:
            painter.fillRect(option.rect, background_color)
            
        if checked:
            check_option.state = option.state | QtGui.QStyle.State_On
        else:
            check_option.state = option.state | QtGui.QStyle.State_Off
            
            
            
        QtGui.QApplication.style().drawControl(QtGui.QStyle.CE_CheckBox,
                                               check_option,
                                               painter)
        
        
        painter.restore()
    
class TextBoolDelegate(CustomDelegate):

    editor = editors.TextBoolEditor
    def __init__(self, parent=None, editable=True, yes='Yes', no='No', color_yes=None, color_no=None, **kwargs):
        CustomDelegate.__init__(self, parent, editable, **kwargs)
        self.yes = yes
        self.no = no
        self.color_no = color_no
        self.color_yes = color_yes

    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        field_attributes = variant_to_pyobject(index.data(Qt.UserRole))
        editable, background_color = True, None
        if field_attributes != ValueLoading:
            editable = field_attributes.get( 'editable', True )
            background_color = field_attributes.get( 'background_color', None )

        rect = option.rect
        
        value = index.model().data(index, Qt.EditRole).toBool()
        font_color = QtGui.QColor()
        if value:
            text = self.yes
            if self.color_yes:
                color = self.color_yes
        else:
            text = self.no
            if self.color_no:
                color = self.color_no
        font_color.setRgb(color.red(), color.green(), color.blue()) 

        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
        else:
            if editable:
                painter.fillRect(option.rect, background_color or option.palette.base())
            else:
                painter.fillRect(option.rect, background_color or option.palette.window())
              
        painter.setPen(font_color.toRgb())
        painter.drawText(
            rect.x() + 2,
            rect.y(),
            rect.width() - 4,
            rect.height(),
            Qt.AlignVCenter | Qt.AlignLeft,
            _(text)
        )
        painter.restore()




#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Classes to layout fields on a form.  These are mostly used for specifying the
form_display attribute in Admin classes, but they can be used on their own as
well.  Form classes can be used recursive.
"""

import logging
logger = logging.getLogger( 'camelot.view.forms' )

from PyQt4 import QtCore, QtGui

from camelot.view.model_thread import gui_function

class Form( list ):
    """Base Form class to put fields on a form.  The base class of a form is
a list.  So the form itself is nothing more than a list of field names or
sub-forms.  A form can thus be manipulated using the list's method such as
append or insert.
    
A form can be converted to a
QT widget by calling its render method.  The base form uses the QFormLayout
to render a form::

    class Admin(EntityAdmin):
      form_display = Form(['title', 'short_description', 'director', 'release_date'])

..image:: /_static/form/form.png
"""

    def __init__( self, content, scrollbars = False, columns = 1  ):
        """
        :param content: an iterable with the field names and forms to render
        :param columns: the number of columns in which to order the fields.

        eg : with 2 columns, the fields ['street', 'city', 'country'] will
        be ordered as :

        +-------------+--------------+
        | street      | city         |
        +-------------+--------------+
        | country     |              |
        +-------------+--------------+

        """
        super(Form, self).__init__( content )
        self._scrollbars = scrollbars
        self._columns = columns

    def get_fields( self ):
        """:return: the fields, visible in this form"""
        return [field for field in self._get_fields_from_form()]

    def _get_fields_from_form( self ):
        for field in self:
            if isinstance( field, Form ):
                for nested_field in  field._get_fields_from_form():
                    yield nested_field
            else:
                assert isinstance( field, ( str, unicode ) )
                yield field;


    def remove_field( self, original_field ):
        """Remove a field from the form, This function can be used to modify
        inherited forms.

        :param original_field: the name of the field to be removed
        :return: True if the field was found and removed
        """
        for c in self:
            if isinstance( c, Form ):
                c.remove_field( original_field )
            if original_field in self:
                self.remove( original_field )
                return True
        return False

    def replace_field( self, original_field, new_field ):
        """Replace a field on this form with another field.  This function can be used to
        modify inherited forms.

        :param original_field : the name of the field to be replace
        :param new_field : the name of the new field
        :return: True if the original field was found and replaced.
        """
        for i, c in enumerate( self ):
            if isinstance( c, Form ):
                c.replace_field( original_field, new_field )
            elif c == original_field:
                self[i] = new_field
                return True
        return False

    def add_field( self, new_field ):
        self.append( new_field )

    def __unicode__( self ):
        return 'Form(%s)' % ( u','.join( unicode( c ) for c in self ) )

    def render_ooxml( self, obj, delegates ):
        """Generator for lines of text in Office Open XML representing this form, using tables
        :param obj: the object or entity that will be rendered
        :param delegates: a dictionary mapping field names to their delegate
        """
        yield '<w:tbl>'
        yield '    <w:tblPr>'
        yield '      <w:tblW w:w="0" w:type="auto"/>'
        yield '      <w:tblBorders>'
        yield '          <w:top w:val="single" w:sz="4" wx:bdrwidth="10" w:space="0" w:color="auto"/>'
        yield '          <w:left w:val="single" w:sz="4" wx:bdrwidth="10" w:space="0" w:color="auto"/>'
        yield '          <w:bottom w:val="single" w:sz="4" wx:bdrwidth="10" w:space="0" w:color="auto"/>'
        yield '          <w:right w:val="single" w:sz="4" wx:bdrwidth="10" w:space="0" w:color="auto"/>'
        yield '          <w:insideH w:val="single" w:sz="4" wx:bdrwidth="10" w:space="0" w:color="auto"/>'
        yield '          <w:insideV w:val="single" w:sz="4" wx:bdrwidth="10" w:space="0" w:color="auto"/>'
        yield '        </w:tblBorders>'
        yield '      <w:tblLook w:val="04A0"/>'
        yield '    </w:tblPr>'
        yield '    <w:tblGrid>'
        yield '      <w:gridCol w:w="4811"/>'
        yield '      <w:gridCol w:w="4811"/>'
        yield '    </w:tblGrid>'
        yield '    <w:tr>'
        for index, field in enumerate(self):
            if index % self._columns == 0 and index != 0:
                yield '    </w:tr><w:tr>'
            yield '<w:tc>'
            yield '  <w:tcPr>'
            yield '    <w:tcW w:w="4811" w:type="dxa"/>'
            yield '    <w:shd w:val="clear" w:color="auto" w:fill="auto"/>'
            yield '  </w:tcPr>'
            yield '<w:p>'
            lines = []
            if isinstance(field, Label):
                lines = field.render_ooxml()
            elif isinstance(field, Form):
                lines = field.render_ooxml( obj, delegates )
            elif isinstance(field, basestring):
                delegate = delegates[field]
                value = getattr(obj, field)
                lines = delegate.render_ooxml(value)
            for line in lines:
                yield line
            yield '</w:p>'
            yield '</w:tc>'
        yield '</w:tr>'
        yield '</w:tbl>'

    @gui_function
    def render( self, widgets, parent = None, toplevel = False):
        """
        :param widgets: a :class:`camelot.view.controls.formview.FormEditors` object
            that is able to create the widgets for this form
        :param parent: the :class:`QtGui.QWidget` in which the form is placed
        :param toplevel: a :keyword:`boolean` indicating if this form is toplevel,
            or a child form of another form.  A toplevel form will be expanding,
            while a non toplevel form is only expanding if it contains other
            expanding elements.

        :return: a QWidget into which the form is rendered
        """
        logger.debug( 'rendering %s' % (self.__class__.__name__) )
        from camelot.view.controls.editors.wideeditor import WideEditor
        form_widget = QtGui.QWidget( parent )
        form_layout = QtGui.QGridLayout()
        # where 1 column in the form is a label and a field, so two columns in the grid
        columns = min(self._columns, len(self))
        # make sure all columns have the same width
        for i in range(columns*2):
            if i%2:
                form_layout.setColumnStretch(i, 1)

        row_span = 1

        class cursor(object):

            def __init__(self):
                self.row = 0
                self.col = 0

            def next_row(self):
                self.row = self.row + 1
                self.col = 0

            def next_col(self):
                self.col = self.col + 2
                if self.col >= columns * 2:
                    self.next_row()

            def next_empty_row(self):
                if self.col!=0:
                    self.next_row()

            def __str__(self):
                return '%s,%s'%(self.row, self.col)

        c = cursor()

        has_vertical_expanding_row = False
        for field in self:
            size_policy = None
            if isinstance( field, Form ):
                c.next_empty_row()
                col_span = 2 * columns
                f = field.render( widgets, parent, False )
                if isinstance( f, QtGui.QLayout ):
                    #
                    # this should maybe be recursive ??
                    #
                    for layout_item_index in range( f.count() ):
                        layout_item = f.itemAt( layout_item_index )
                        layout_item_widget = layout_item.widget()
                        if layout_item_widget and layout_item_widget.sizePolicy().verticalPolicy() == QtGui.QSizePolicy.Expanding:
                            has_vertical_expanding_row = True
                    form_layout.addLayout( f, c.row, c.col, row_span, col_span )
                else:
                    form_layout.addWidget( f, c.row, c.col, row_span, col_span )
                    size_policy = f.sizePolicy()
                c.next_row()
            else:
                editor = widgets.create_editor( field, form_widget )
                if editor != None:
                    if isinstance( editor, ( WideEditor, ) ):
                        c.next_empty_row()
                        col_span = 2 * columns
                        label = widgets.create_label( field, editor, form_widget )
                        if label:
                            form_layout.addWidget( label, c.row, c.col, row_span, col_span )
                            c.next_row()
                        form_layout.addWidget( editor, c.row, c.col, row_span, col_span )
                        c.next_row()
                    else:
                        col_span = 1
                        label = widgets.create_label( field, editor, form_widget )
                        if label:
                            form_layout.addWidget( label, c.row, c.col, row_span, col_span )
                        form_layout.addWidget( editor, c.row, c.col + 1, row_span, col_span )
                        c.next_col()
                    size_policy = editor.sizePolicy()
                else:
                    logger.warning('ProgrammingError : widgets should contain a widget for field %s'%unicode(field))
            if size_policy and size_policy.verticalPolicy() == QtGui.QSizePolicy.Expanding:
                has_vertical_expanding_row = True

        if (not has_vertical_expanding_row) and toplevel and form_layout.rowCount():
            form_layout.setRowStretch( form_layout.rowCount(), 1 )

        # fix embedded forms
        if not toplevel:
            form_layout.setContentsMargins( 0, 0, 0, 0 )
  
        if toplevel or has_vertical_expanding_row:
            form_widget.setSizePolicy( QtGui.QSizePolicy.Expanding,
                                       QtGui.QSizePolicy.Expanding )
        form_widget.setLayout( form_layout )

        if self._scrollbars:
            scroll_area = QtGui.QScrollArea( parent )
            # we should inherit parent's background color
            scroll_area.setWidget( form_widget )
            scroll_area.setWidgetResizable( True )
            scroll_area.setFrameStyle( QtGui.QFrame.NoFrame )
            return scroll_area

        logger.debug( 'end rendering %s' % self.__class__.__name__ )
        
        return form_widget


class Label( Form ):
    """Render a label with a QLabel"""

    def __init__( self, label, alignment='left', style=None):
        """
        :param label : string to be displayed in the label
        :param alignment : alignment of text in the label. values that make sense 'left', 'right' or 'center'
        :param style : string of cascading stylesheet instructions
        """
        super( Label, self ).__init__( [] )
        self.label = label
        self.alignment = alignment
        self.style = style

    def render_ooxml( self ):
        """Generator for label text in Office Open XML representing this form"""
        yield '<w:r>'
        yield '  <w:t>%s</w:t>' % self.label
        yield '</w:r>'


    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        from PyQt4 import QtGui
        if self.style:
            widget = QtGui.QLabel( '<p align="%s" style="%s">%s</p>' % (self.alignment, self.style,unicode(self.label)) )
        else:
            widget = QtGui.QLabel( '<p align="%s">%s</p>' % (self.alignment,unicode(self.label)) )
        return widget

class DelayedTabWidget( QtGui.QTabWidget ):
    """Helper class for :class:`TabForm` to delay the creation of tabs to
the moment the tab is shown.
    """
    
    def __init__( self,  widgets, tabs, parent = None ):
        super( DelayedTabWidget, self ).__init__( parent )
        self._widgets = widgets
        self._forms = []
        #
        # keep track for each of the tabs wether they are expanding,
        #
        self._vertical_expanding = [False] * len(tabs)
        #
        # set dummy tab widgets
        #
        for tab_label, tab_form in tabs:
            self._forms.append( tab_form )
            tab_widget = QtGui.QWidget( self )
            self.addTab( tab_widget, unicode(tab_label) )
        #
        # render the first tab and continue rendering until we have
        # a tab with an expanding size policy, because then we know
        # the widget itself should be expanding, and we can delay
        # the rendering of the other tabs
        #
        for i in range( len(tabs) ):
            self.render_tab( i )
            if sum(self._vertical_expanding):
                self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
                #
                # if one of the tabs is expanding, the others should have spacer
                # items to stretch
                #
                for j, vertical_expanding_of_widget in zip(range(i), self._vertical_expanding):
                    if vertical_expanding_of_widget == False:
                        tab_widget = self.widget( j )
                        tab_widget.layout().addStretch( 1 )                
                break
        self.currentChanged.connect( self.render_tab )
            
    @QtCore.pyqtSlot( int )
    def render_tab( self, index ):
        """
        Render the tab at index
        """
        tab_widget = self.widget( index )
        layout = tab_widget.layout()
        if layout != None:
            # this tab has been rendered before
            return
        layout = QtGui.QVBoxLayout( tab_widget )
        tab_form = self._forms[index]
        tab_form_widget = tab_form.render( self._widgets, tab_widget, False )
        layout.addWidget( tab_form_widget )
        tab_widget.setLayout( layout )
        size_policy = tab_form_widget.sizePolicy()
        if size_policy.verticalPolicy() == QtGui.QSizePolicy.Expanding:
            self._vertical_expanding[index] = True
        else:
            self._vertical_expanding[index] = False
        #
        # if other widgets are expanding, and this one isn't, add some stretch
        #
        if self._vertical_expanding[index] == False and sum(self._vertical_expanding):
            tab_widget.layout().addStretch( 1 )

class TabForm( Form ):
    """
Render forms within a QTabWidget::

    from = TabForm([('First tab', ['title', 'short_description']),
                    ('Second tab', ['director', 'release_date'])])

.. image:: /_static/form/tab_form.png
    """

    NORTH = 'North'
    SOUTH = 'South'
    WEST = 'West'
    EAST = 'East'

    def __init__( self, tabs, position=NORTH ):
        """
        :param tabs: a list of tuples of (tab_label, tab_form)
        :param position: the position of the tabs with respect to the pages
        """
        assert isinstance( tabs, list )
        assert position in [self.NORTH, self.SOUTH, self.WEST, self.EAST]
        self.position = position
        for tab in tabs:
            assert isinstance( tab, tuple )
        self.tabs = [( tab_label, structure_to_form( tab_form ) ) for tab_label, tab_form in tabs]
        super( TabForm, self ).__init__( sum( ( tab_form.get_fields()
                                      for tab_label, tab_form in self.tabs ), [] ) )

    def __unicode__( self ):
        return 'TabForm { %s\n        }' % ( u'\n  '.join( '%s : %s' % ( label, unicode( form ) ) for label, form in self.tabs ) )

    def add_tab_at_index( self, tab_label, tab_form, index ):
        """Add a tab to the form at the specified index

  :param tab_label: the name to the tab
  :param tab_form: the form to display in the tab or a list of field names.
  :param index: the position of tab in the tabs list.
        """
        tab_form = structure_to_form( tab_form )
        self.tabs.insert( index, ( tab_label, tab_form ) )
        self.extend( [tab_form] )

    def add_tab( self, tab_label, tab_form ):
        """Add a tab to the form

    :param tab_label: the name of the tab
    :param tab_form: the form to display in the tab or a list of field names.
        """
        tab_form = structure_to_form( tab_form )
        self.tabs.append( ( tab_label, tab_form ) )
        self.extend( [tab_form] )

    def get_tab( self, tab_label ):
        """Get the tab form of associated with a tab_label, use this function to
        modify the underlying tab_form in case of inheritance

    :param tab_label : a label of a tab as passed in the construction method
    :return: the tab_form corresponding to tab_label
        """
        for label, form in self.tabs:
            if label == tab_label:
                return form

    def replace_field( self, original_field, new_field ):
        super(TabForm, self).replace_field( original_field, new_field )
        for _label, form in self.tabs:
            if form.replace_field( original_field, new_field ):
                return True
        return False

    def remove_field( self, original_field ):
        super(TabForm, self).remove_field( original_field )
        for _label, form in self.tabs:
            if form.remove_field( original_field ):
                return True
        return False

    def _get_fields_from_form( self ):
        for _label, form in self.tabs:
            for field in form._get_fields_from_form():
                yield field

    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        logger.debug( 'rendering %s' % self.__class__.__name__ )
        widget = DelayedTabWidget( widgets, self.tabs, parent )
        widget.setTabPosition( getattr(QtGui.QTabWidget, self.position) )                
        return widget

class HBoxForm( Form ):
    """
  Render different forms in a horizontal box::

    form = forms.HBoxForm([['title', 'short_description'], ['director', 'release_date']])

  .. image:: /_static/form/hbox_form.png

  """

    def __init__( self, columns, scrollbars=False ):
        """:param columns: a list of forms to display in the different columns
        of the horizontal box"""
        assert isinstance( columns, list )
        self.columns = [structure_to_form( col ) for col in columns]
        super( HBoxForm, self ).__init__( sum( ( column_form.get_fields()
                                      for column_form in self.columns ), [] ), scrollbars=scrollbars )

    def __unicode__( self ):
        return 'HBoxForm [ %s\n         ]' % ( '         \n'.join( [unicode( form ) for form in self.columns] ) )

    def replace_field( self, original_field, new_field ):
        for form in self.columns:
            if form.replace_field( original_field, new_field ):
                return True
        return False

    def _get_fields_from_form( self ):
        for form in self.columns:
            for field in form._get_fields_from_form():
                yield field

    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        logger.debug( 'rendering %s' % self.__class__.__name__ )
        widget = QtGui.QWidget( parent )
        form_layout = QtGui.QHBoxLayout()
        for form in self.columns:
            f = form.render( widgets, widget, False )
            if isinstance( f, QtGui.QLayout ):
                form_layout.addLayout( f )
            else:
                form_layout.addWidget( f )
        widget.setLayout( form_layout )
        return widget

class VBoxForm( Form ):
    """
  Render different forms or widgets in a vertical box::

    form = forms.VBoxForm([['title', 'short_description'], ['director', 'release_date']])

  .. image:: /_static/form/vbox_form.png
  """

    def __init__( self, rows ):
        """:param rows: a list of forms to display in the different columns
        of the horizontal box
        """
        assert isinstance( rows, list )
        self.rows = [structure_to_form( row ) for row in rows]
        super( VBoxForm, self ).__init__( sum( ( row_form.get_fields() for row_form in self.rows ), [] ) )

    def replace_field( self, original_field, new_field ):
        for form in self.rows:
            if form.replace_field( original_field, new_field ):
                return True
        return False

    def _get_fields_from_form( self ):
        for form in self.rows:
            for field in form._get_fields_from_form():
                yield field

    def __unicode__( self ):
        return 'VBoxForm [ %s\n         ]' % ( '         \n'.join( [unicode( form ) for form in self.rows] ) )

    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        logger.debug( 'rendering %s' % self.__class__.__name__ )
        widget = QtGui.QWidget( parent )
        form_layout = QtGui.QVBoxLayout()
        for form in self.rows:
            f = form.render( widgets, widget, False )
            if isinstance( f, QtGui.QLayout ):
                form_layout.addLayout( f )
            else:
                form_layout.addWidget( f )
        widget.setLayout( form_layout )
        return widget

class ColumnSpan( Form ):
    
    def __init__(self, field=None, num=2):
        self.num = num
        self.field = field
        super( ColumnSpan, self ).__init__( [field] )

class GridForm( Form ):
    """Put different fields into a grid, without a label.  Row or column labels can be added
  using the Label form::

    GridForm([['title', 'short_description'], ['director','release_date']])

  .. image:: /_static/form/grid_form.png
  """

    def __init__( self, grid, nomargins = False ):
        """:param grid: A list for each row in the grid, containing a list with all fields that should be put in that row
        """
        assert isinstance( grid, list )
        self._grid = grid
        self._nomargins = nomargins
        fields = []
        for row in grid:
            assert isinstance( row, list )
            fields.extend( row )
        super( GridForm, self ).__init__( fields )

    def append_row(self, row):
        """:param row: the list of fields that should come in the additional row
        use this method to modify inherited grid forms"""
        assert isinstance( row, list )
        self.extend(row)
        self._grid.append(row)

    def append_column(self, column):
        """:param column: the list of fields that should come in the additional column
        use this method to modify inherited grid forms"""
        assert isinstance( column, list )
        self.extend(column)
        for row, additional_field in zip(self._grid, column):
            row.append(additional_field)

    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        widget = QtGui.QWidget( parent )
        grid_layout = QtGui.QGridLayout()
        for i, row in enumerate( self._grid ):
            skip = 0
            for j, field in enumerate( row ):
                num = 1
                if isinstance( field, ColumnSpan ):
                    num = field.num
                    field = field.field

                if isinstance( field, Form ):
                    grid_layout.addWidget( field.render( widgets, parent ), i, j + skip, 1, num )
                    skip += num - 1
                else:
                    editor = widgets.create_editor( field, widget )
                    grid_layout.addWidget( editor, i, j + skip, 1, num )
                    skip += num - 1

        widget.setLayout( grid_layout )
        if not toplevel:
            grid_layout.setContentsMargins( 0, 0, 0, 0 )

        return widget

class WidgetOnlyForm( Form ):
    """Renders a single widget without its label, typically a one2many widget"""

    def __init__( self, field ):
        assert isinstance( field, ( str, unicode ) )
        super( WidgetOnlyForm, self ).__init__( [field] )

    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        logger.debug( 'rendering %s' % self.__class__.__name__ )
        editor = widgets.create_editor( self.get_fields()[0], parent )
        return editor

class GroupBoxForm( Form ):
    """
  Renders a form within a QGroupBox::

    class Admin(EntityAdmin):
      form_display = GroupBoxForm('Movie', ['title', 'short_description'])

  .. image:: /_static/form/group_box_form.png
  """

    def __init__( self, title, content, scrollbars=None, min_width=None, min_height=None, columns=1 ):
        self.title = title
        self.min_width = min_width
        self.min_height = min_height
        if isinstance(content, Form):
            content = [content]
        Form.__init__( self, content, scrollbars, columns=columns )

    @gui_function
    def render( self, widgets, parent = None, toplevel = False ):
        widget = QtGui.QGroupBox( unicode(self.title), parent )
        layout = QtGui.QVBoxLayout()
        if self.min_width and self.min_height:
            widget.setMinimumSize ( self.min_width, self.min_height )
        widget.setLayout( layout )
        form = Form.render( self, widgets, widget, False )
        layout.addWidget( form )
        return widget

def structure_to_form( structure ):
    """Convert a python data structure to a form, using the following rules :

   * if structure is an instance of Form, return structure
   * if structure is a list, create a Form from this list

  This function is mainly used in the Admin class to construct forms out of
  the form_display attribute
    """
    if isinstance( structure, Form ):
        return structure
    return Form( structure )





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from PyQt4 import QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.core import constants
from camelot.core.utils import variant_to_pyobject
from camelot.view.proxy import ValueLoading

class FloatDelegate( CustomDelegate ):
    """Custom delegate for float values"""

    __metaclass__ = DocumentationMetaclass

    editor = editors.FloatEditor

    def __init__( self,
                 minimum=constants.camelot_minfloat,
                 maximum=constants.camelot_maxfloat,
                 parent=None,
                 unicode_format=None,
                 **kwargs ):
        CustomDelegate.__init__(self,
                                parent=parent,
                                minimum=minimum, maximum=maximum,
                                **kwargs )                   
        self.minimum = minimum
        self.maximum = maximum
        self.unicode_format = unicode_format
        self._locale = QtCore.QLocale()

    def paint( self, painter, option, index ):
        painter.save()
        self.drawBackground(painter, option, index)
        value = variant_to_pyobject(index.model().data(index, Qt.EditRole))
        field_attributes = variant_to_pyobject( index.model().data( index, Qt.UserRole ) )

        if field_attributes == ValueLoading:
            precision = 2
        else:
            precision = field_attributes.get('precision', 2)
            
        if value in (None, ValueLoading):
            value_str = ''
        elif self.unicode_format:
            value_str = self.unicode_format(value)
        else:
            value_str = unicode( self._locale.toString( float(value), 
                                                        'f', 
                                                        precision ) )

        self.paint_text( painter, option, index, value_str, horizontal_align=Qt.AlignRight )
        painter.restore()




#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import functools
import logging
logger = logging.getLogger('camelot.view.mainwindow')

#
# Dummy imports to fool the windows installer and force
# it to include the right packages
#
# from sqlalchemy.databases import sqlite
# import sqlite3

from PyQt4.QtCore import Qt
from PyQt4 import QtGui, QtCore

from camelot.view.action import ActionFactory
from camelot.view.controls.navpane2 import NavigationPane
from camelot.view.controls.progress_dialog import ProgressDialog
from camelot.view.model_thread import post

from camelot.core.utils import ugettext as _

def addActions(target, actions):
    """add action objects to menus, menubars, and toolbars
    if action is None, add a separator.
    """
    for action in actions:
        if action is None:
            target.addSeparator()
        else:
            target.addAction(action)

class MainWindow(QtGui.QMainWindow):
    """Main window GUI"""

    def __init__(self, app_admin, parent=None):
        from workspace import DesktopWorkspace
        logger.debug('initializing main window')
        QtGui.QMainWindow.__init__(self, parent)

        self.app_admin = app_admin
        logger.debug('setting up workspace')
        self.workspace = DesktopWorkspace(app_admin, self)

        logger.debug('setting child windows dictionary')

        logger.debug('setting central widget to our workspace')
        self.setCentralWidget(self.workspace)

        self.workspace.view_activated_signal.connect(self.updateMenus)
        self.workspace.change_view_mode_signal.connect( self.change_view_mode )
        self.workspace.last_view_closed_signal.connect( self.unmaximize_view )

        logger.debug('creating navigation pane')
        self.createNavigationPane()

        logger.debug('creating all the required actions')
        self.createActions()

        logger.debug('creating the menus')
        self.createMenus()

        logger.debug('creating the toolbars')
        self.createToolBars()

        logger.debug('creating status bar')
        self.createStatusBar()

        logger.debug('updating menus')
        self.updateMenus()

        logger.debug('reading saved settings')
        self.readSettings()
        
        windowtitle = self.app_admin.get_name()
        logger.debug('setting up window title: %s' % windowtitle)
        self.setWindowTitle(windowtitle)
        self.app_admin.title_changed_signal.connect( self.setWindowTitle )

        #QtCore.QTimer.singleShot(0, self.doInitialization)
        logger.debug('initialization complete')
        
    def about(self):
        logger.debug('showing about message box')
        abtmsg = self.app_admin.get_about()
        QtGui.QMessageBox.about(self, _('About'), _(abtmsg))
        logger.debug('about message closed')

    def whats_new(self):
        widget = self.app_admin.get_whats_new()
        if widget:
            widget.exec_()

    def affiliated_website(self):
        from PyQt4.QtGui import QDesktopServices
        url = self.app_admin.get_affiliated_url()
        if url:
            QDesktopServices.openUrl(url)

    def remote_support(self):
        from PyQt4.QtGui import QDesktopServices
        url = self.app_admin.get_remote_support_url()
        if url:
            QDesktopServices.openUrl(url)

    @QtCore.pyqtSlot()
    def unmaximize_view(self):
        self.navpane.show()
        self.menuBar().show()

    @QtCore.pyqtSlot()
    def change_view_mode(self):
        if self.menuBar().isHidden():
            self.navpane.show()
            self.menuBar().show()
        else:
            self.navpane.hide()
            self.menuBar().hide()

    def readSettings(self):
        settings = QtCore.QSettings()
        self.restoreGeometry(settings.value('geometry').toByteArray())
        # Don't restore state, since it messes up the toolbar if stuff was
        # added
        # self.restoreState(settings.value('state').toByteArray())

    def writeSettings(self):
        logger.debug('writing application settings')
        settings = QtCore.QSettings()
        settings.setValue('geometry', QtCore.QVariant(self.saveGeometry()))
        #settings.setValue('state', QtCore.QVariant(self.saveState()))
        logger.debug('settings written')

    def display_exception_message_box(self, exc_info):
        from controls.exception import model_thread_exception_message_box
        model_thread_exception_message_box(exc_info)

    def runAction(self, name, callable):
        progress = QtGui.QProgressDialog(_('Please wait'), QtCore.QString(), 0, 0)
        progress.setWindowTitle(name)
        progress.show()
        post(
            callable,
            progress.close,
            exception=self.display_exception_message_box
        )

    def createActions(self):
        self.backupAct = ActionFactory.backup(self, self.backup)
        self.restoreAct = ActionFactory.restore(self, self.restore)
        self.pageSetupAct = ActionFactory.page_setup(self, self.pageSetup)
        self.previewAct = ActionFactory.print_preview(self, self.previewDoc)
        self.exitAct= ActionFactory.exit(self, slot=self.close)
        self.copyAct = ActionFactory.copy(self, slot=self.copy)
        self.selectAllAct = ActionFactory.select_all(self, slot=self.select_all)
        self.separatorAct = QtGui.QAction(self)
        self.separatorAct.setSeparator(True)
        self.aboutAct = ActionFactory.about(self, slot=self.about)
        self.whats_new_action = ActionFactory.whats_new(self, slot=self.whats_new)
        self.affiliated_website_action = ActionFactory.affiliated_website(self, slot=self.affiliated_website)
        self.remote_support_action =ActionFactory.remote_support(self, slot=self.remote_support)
        self.helpAct = ActionFactory.help(self, slot=self.help)
        self.newAct = ActionFactory.new(self, slot=self.new)
        self.deleteAct = ActionFactory.delete(self, slot=self.delete)
        self.viewFirstAct = ActionFactory.view_first(self, self.viewFirst)
        self.viewLastAct = ActionFactory.view_last(self, self.viewLast)
        self.viewNextAct = ActionFactory.view_next(self, self.viewNext)
        self.viewPreviousAct = ActionFactory.view_previous(self, self.viewPrevious)
        self.viewFirstAct.setIconVisibleInMenu(False)
        self.viewLastAct.setIconVisibleInMenu(False)
        self.viewNextAct.setIconVisibleInMenu(False)
        self.viewPreviousAct.setIconVisibleInMenu(False)
        self.updateValueAct = ActionFactory.update_values(self, slot = self.updateValue)
        self.mergeDocumentAct = ActionFactory.merge_document(self, slot = self.merge_document)
        self.exportToExcelAct = ActionFactory.export_excel(self, slot=self.exportToExcel)
        self.exportToWordAct = ActionFactory.export_word(self, slot=self.exportToWord)
        self.exportToMailAct = ActionFactory.export_mail(self, slot=self.exportToMail)
        self.importFromFileAct = ActionFactory.import_file(self, slot=self.importFromFile)
        self.sessionRefreshAct = ActionFactory.refresh(self, slot=self.refresh_session)

    # QAction slots and methods implementations

    def refresh_session(self):
        from elixir import session
        from camelot.core.orm import refresh_session
        post( functools.update_wrapper( functools.partial( refresh_session, session ), refresh_session ) )
        self.workspace.refresh()

    def help(self):
        #
        # Import WebKit as late as possible, since it's the largest
        # part of the QT Library (15 meg on Ubuntu linux)
        #
        from PyQt4 import QtWebKit
        TOP_LEVEL = None
        self.view = QtWebKit.QWebView(TOP_LEVEL)
        #print self.app_admin.get_help_url()
        #print self.app_admin.get_help_base()
        #index_file = open(self.app_admin.get_help_url(),'r')
        #self.view.setHtml (index_file.read(), self.app_admin.get_help_base())
        self.view.load(self.app_admin.get_help_url())
        self.view.setWindowTitle(_('Help Browser'))
        self.view.setWindowIcon(self.helpAct.icon())
        self.view.show()

    def backup(self):
        self.app_admin.backup(self)

    def restore(self):
        self.app_admin.restore(self)

    def saveAs(self):
        pass

    def copy(self):
        self.activeMdiChild().copy_selected_rows()

    def select_all(self):
        self.activeMdiChild().select_all_rows()

    def previewDoc(self):
        active = self.activeMdiChild()
        from camelot.admin.action import Action, GuiContext
        from camelot.view.action_steps import PrintHtml

        class PrintPreviewAction( Action ):
            
            def model_run( self, model_context ):
                yield PrintHtml( active.to_html() )

        action = PrintPreviewAction()
        action.gui_run( GuiContext() )

    def new(self):
        self.activeMdiChild().newRow()

    def delete(self):
        self.activeMdiChild().deleteSelectedRows()

    def pageSetup(self):
        pass

    def viewFirst(self):
        """selects view's first row"""
        active = self.activeMdiChild()
        active.viewFirst()

    def viewLast(self):
        """selects view's last row"""
        active = self.activeMdiChild()
        active.viewLast()

    def viewNext(self):
        """selects view's next row"""
        active = self.activeMdiChild()
        active.viewNext()

    def viewPrevious(self):
        """selects view's previous row"""
        active = self.activeMdiChild()
        active.viewPrevious()

    def updateValue(self):
        from camelot.view.wizard.update_value import UpdateValueWizard

        admin = self.activeMdiChild().get_admin()
        selection_getter = self.activeMdiChild().get_selection_getter()
        wizard = UpdateValueWizard(admin=admin, selection_getter=selection_getter)
        wizard.exec_()

    def merge_document(self):
        """Run the merge document wizard on the selection in the current
        table view"""
        from camelot.view.wizard.merge_document import MergeDocumentWizard

        selection_getter = self.activeMdiChild().get_selection_getter()
        wizard = MergeDocumentWizard(selection_getter=selection_getter)
        wizard.exec_()

    def exportToExcel(self):
        """creates an excel file from the view"""
        widget = self.activeMdiChild()
        d = ProgressDialog(_('Please wait'))
        post(widget.export_to_excel, d.finished, d.exception)
        d.exec_()

    def exportToWord(self):
        """Use windows COM to export the active child window to MS word,
        by using its to_html function"""
        widget = self.activeMdiChild()
        d = ProgressDialog(_('Please wait'))
        post(widget.export_to_word, d.finished, d.exception)
        d.exec_()

    def exportToMail(self):
        widget = self.activeMdiChild()
        d = ProgressDialog(_('Please wait'))
        post(widget.export_to_mail, d.finished, d.exception)
        d.exec_()

    def importFromFile(self):
        self.activeMdiChild().importFromFile()

    def createMenus(self):

        self.file_menu = self.menuBar().addMenu(_('&File'))
        addActions(self.file_menu, (
            #self.closeAct,
            None,
            self.backupAct,
            self.restoreAct,
            None,
            self.pageSetupAct,
            self.previewAct,
            None
        ))

        self.exportMenu = QtGui.QMenu(_('Export To'))
        addActions(self.exportMenu, (
            self.exportToExcelAct,
            self.exportToWordAct,
            self.exportToMailAct,
        ))
        self.file_menu.addMenu(self.exportMenu)

        self.importMenu = QtGui.QMenu(_('Import From'))
        addActions(self.importMenu, (self.importFromFileAct,))
        self.file_menu.addMenu(self.importMenu)

        addActions(self.file_menu, (None, self.exitAct))

        self.editMenu = self.menuBar().addMenu(_('&Edit'))

        addActions(self.editMenu, (self.copyAct,
                                   self.selectAllAct,
                                   self.updateValueAct,
                                   self.mergeDocumentAct))

        self.viewMenu = self.menuBar().addMenu(_('View'))
        addActions(self.viewMenu, (self.sessionRefreshAct,))
        gotoMenu = self.viewMenu.addMenu(_('Go To'))
        addActions(gotoMenu, (
            self.viewFirstAct,
            self.viewPreviousAct,
            self.viewNextAct,
            self.viewLastAct
        ))

        self.menuBar().addSeparator()

        self.helpMenu = self.menuBar().addMenu(_('&Help'))

        help_menu_actions = [self.helpAct, self.aboutAct]
        if self.app_admin.get_whats_new():
            help_menu_actions.append(self.whats_new_action)
        if self.app_admin.get_affiliated_url():
            help_menu_actions.append(self.affiliated_website_action)
        if self.app_admin.get_remote_support_url():
            help_menu_actions.append(self.remote_support_action)
        addActions(self.helpMenu, help_menu_actions )

    def updateMenus(self):
        """Toggle the status of the menus, depending on the active view"""
        active_view = (self.workspace.active_view() is not None)
        self.backupAct.setEnabled(True)
        self.restoreAct.setEnabled(True)

        self.pageSetupAct.setEnabled(active_view)
        self.previewAct.setEnabled(active_view)

        self.newAct.setEnabled(active_view)
        self.deleteAct.setEnabled(active_view)
        self.copyAct.setEnabled(active_view)
        self.viewFirstAct.setEnabled(active_view)
        self.viewPreviousAct.setEnabled(active_view)
        self.viewNextAct.setEnabled(active_view)
        self.viewLastAct.setEnabled(active_view)

        self.exportToWordAct.setEnabled(active_view)
        self.exportToExcelAct.setEnabled(active_view)
        self.exportToMailAct.setEnabled(active_view)

        self.importFromFileAct.setEnabled(active_view)

        self.separatorAct.setVisible(active_view)

    def get_tool_bar(self):
        return self.tool_bar

    def createToolBars(self):
        #
        # All actions are put in one toolbar, to ease unit testing and
        # generation of screenshots
        #
        self.tool_bar = self.addToolBar(_('Toolbar'))
        self.tool_bar.setObjectName('ToolBar')
        self.tool_bar.setMovable(False)
        self.tool_bar.setFloatable(False)
        addActions(self.tool_bar, (
            self.newAct,
            self.copyAct,
            self.deleteAct,
            self.viewFirstAct,
            self.viewPreviousAct,
            self.viewNextAct,
            self.viewLastAct
        ))

        addActions(self.tool_bar, (
            self.exportToExcelAct,
            self.exportToWordAct,
            self.exportToMailAct,
        ))

        addActions(self.tool_bar, (self.previewAct,))

        addActions(self.tool_bar, (self.helpAct,))

    # Navigation Pane
    def createNavigationPane(self):
        self.navpane = NavigationPane(
            self.app_admin,
            workspace=self.workspace,
            parent=self
        )
        self.addDockWidget(Qt.LeftDockWidgetArea, self.navpane)

    def activeMdiChild(self):
        return self.workspace.active_view()

    # Statusbar
    def createStatusBar(self):
        from controls.statusbar import StatusBar
        statusbar = StatusBar(self)
        self.setStatusBar(statusbar)
        statusbar.showMessage(_('Ready'), 5000)

    def closeEvent(self, event):
        from camelot.view.model_thread import get_model_thread
        model_thread = get_model_thread()
        self.workspace.close_all_views()
        self.writeSettings()
        logger.info( 'closing mainwindow' )
        model_thread.stop()
        super( MainWindow, self ).closeEvent( event )
        QtCore.QCoreApplication.exit(0)



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from camelot.model import metadata
from camelot.admin.action import Action
import camelot.types
from elixir.entity import Entity
from elixir.options import using_options
from elixir.fields import Field
from sqlalchemy.types import Unicode, INT
"""Set of classes to enable internationalization of the user interface"""

__metadata__ = metadata

from camelot.admin.entity_admin import EntityAdmin
from camelot.view.art import Icon
from camelot.core.utils import ugettext_lazy as _
from camelot.view.utils import default_language

import logging
logger = logging.getLogger( 'camelot.model.i18n' )

def tr( source ):
    from PyQt4 import QtCore
    language = unicode( QtCore.QLocale().name() )
    return Translation.translate_or_register( source, language )

class ExportAsPO( Action ):

    verbose_name = _('PO Export')
    icon = Icon('tango/16x16/actions/document-save.png')

    def model_run( self, model_context ):
        from camelot.view.action_steps import SelectFile
        select_file = SelectFile()
        select_file.existing = False
        filenames = yield select_file
        for filename in filenames:
            file = open(filename, 'w')
            for translation in model_context.get_collection():
                file.write( (u'msgid  "%s"\n'%translation.source).encode('utf-8') )
                file.write( (u'msgstr "%s"\n\n'%translation.value).encode('utf-8') )
                
        
class Translation( Entity ):
    using_options( tablename = 'translation' )
    language = Field( camelot.types.Language, index = True )
    source = Field( Unicode( 500 ), index = True )
    # value needs to be indexed as well, because when starting up we
    # want to load only the translations that have a value specified
    value = Field( Unicode( 500 ), index = True )
    cid = Field( INT(), default = 0, index = True )
    uid = Field( INT(), default = 0, index = True )

    # cache, to prevent too much of the same sql queries
    _cache = dict()

    class Admin( EntityAdmin ):
        verbose_name_plural = _( 'Translations' )
        form_size = ( 700, 150 )
        section = 'configuration'
        list_display = ['source', 'language', 'value', 'uid']
        list_filter = ['language']
        list_actions = [ExportAsPO()]
        field_attributes = {'language':{'default':default_language}}

    @classmethod
    def translate( cls, source, language ):
        """Translate source to language, return None if no translation is found"""
        if source:
            key = ( source, language )
            if key in cls._cache:
                return cls._cache[key]
            translation = cls.query.filter_by( source = unicode( source ), language = language ).filter( Translation.uid != 0 ).first()
            if translation:
                cls._cache[key] = translation.value
                return translation.value
            return None
        return ''

    @classmethod
    def translate_or_register( cls, source, language ):
        """Translate source to language, if no translation is found, register the
        source as to be translated and return the source"""
        if source:
            source = unicode( source )
            translation = cls.translate( source, language )
            if not translation:
                if not cls.query.filter_by( source = source, language = language ).first():
                    if ( source, language ) not in cls._cache:
                        from elixir import session
                        registered_translation = Translation( source = source, language = language )
                        cls._cache[( source, language )] = source
                        session.flush( [registered_translation] )
                        logger.debug( 'registed %s with id %s' % ( source, registered_translation.id ) )
                return source
            return translation
        return ''





from camelot.admin.validator.entity_validator import EntityValidator
from camelot.admin.entity_admin import EntityAdmin

class PersonValidator(EntityValidator):

    def objectValidity(self, entity_instance):
        messages = super(PersonValidator,self).objectValidity(entity_instance)
        if (not entity_instance.first_name) or (len(entity_instance.first_name) < 3):
            messages.append("A person's first name should be at least 2 characters long")
        return messages
    
class Admin(EntityAdmin):
    verbose_name = 'Person'
    list_display = ['first_name', 'last_name']
    validator = PersonValidator    



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from PyQt4 import QtGui
from PyQt4 import QtCore
from PyQt4.QtCore import Qt
from wideeditor import WideEditor
from customeditor import CustomEditor
from camelot.view.art import Icon

class CustomTextEdit(QtGui.QTextEdit):
    """
    A TextEdit editor that sends editingFinished events 
    when the text was changed and focus is lost.
    """

    editingFinished = QtCore.pyqtSignal()
    receivedFocus = QtCore.pyqtSignal()
    
    def __init__(self, parent):
        super(CustomTextEdit, self).__init__(parent)
        self._changed = False
        self.setTabChangesFocus( True )
        self.textChanged.connect( self._handle_text_changed )

    def focusInEvent(self, event):
        super(CustomTextEdit, self).focusInEvent( event )
        self.receivedFocus.emit()

    def focusOutEvent(self, event):
        if self._changed:
            self.editingFinished.emit()
        super(CustomTextEdit, self).focusOutEvent( event )

    def _handle_text_changed(self):
        self._changed = True

    def setTextChanged(self, state=True):
        self._changed = state

    def setHtml(self, html):
        QtGui.QTextEdit.setHtml(self, html)
        self._changed = False
                
class RichTextEditor(CustomEditor, WideEditor):

    def __init__(self, 
                 parent = None, 
                 field_name = 'richtext',
                 **kwargs):
        CustomEditor.__init__(self, parent)
        self.setObjectName( field_name )
        self.layout = QtGui.QVBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins( 0, 0, 0, 0)
        self.setSizePolicy( QtGui.QSizePolicy.Expanding,
                            QtGui.QSizePolicy.Expanding )

        self.textedit = CustomTextEdit(self)

        self.initToolbar() # Has to be invoked before the connect's below.
        self.toolbar.hide() # Should only be visible when textedit is focused.
        
        self.textedit.editingFinished.connect(self.emit_editing_finished)
        self.textedit.receivedFocus.connect(self.toolbar.show)
        self.textedit.setAcceptRichText(True)
        
        # Layout
        self.layout.addWidget(self.toolbar)
        self.layout.addWidget(self.textedit)
        self.setLayout(self.layout)

        # Format
        self.textedit.setFontWeight(QtGui.QFont.Normal)
        self.textedit.setFontItalic(False)
        self.textedit.setFontUnderline(False)
        #self.textedit.setFocus(Qt.OtherFocusReason)
        self.update_alignment()
        self.textedit.currentCharFormatChanged.connect(self.update_format)
        self.textedit.cursorPositionChanged.connect(self.update_text)

    @QtCore.pyqtSlot()
    def emit_editing_finished(self):
        if self.textedit._changed:
            self.editingFinished.emit()

    def set_editable(self, editable):
        self.textedit.setReadOnly(not editable)

    def set_field_attributes(self, editable=True, background_color=None, **kwargs):
        self.set_editable(editable)
        self.set_background_color(background_color)

    def initToolbar(self):
        self.toolbar = QtGui.QToolBar(self)
        self.toolbar.setOrientation(Qt.Horizontal)
        self.toolbar.setContentsMargins(0, 0, 0, 0)

        self.bold_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-text-bold.png').getQIcon()
        self.bold_button.setIcon(icon)
        self.bold_button.setAutoRaise(True)
        self.bold_button.setCheckable(True)
        self.bold_button.setFocusPolicy( Qt.ClickFocus )
        self.bold_button.setMaximumSize(QtCore.QSize(20, 20))
        self.bold_button.setShortcut(QtGui.QKeySequence('Ctrl+B'))
        self.bold_button.setToolTip('Bold')
        self.bold_button.clicked.connect(self.set_bold)

        self.italic_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-text-italic.png').getQIcon()
        self.italic_button.setIcon(icon)
        self.italic_button.setAutoRaise(True)
        self.italic_button.setCheckable(True)
        self.italic_button.setFocusPolicy( Qt.ClickFocus )
        self.italic_button.setMaximumSize(QtCore.QSize(20, 20))
        self.italic_button.setShortcut(QtGui.QKeySequence('Ctrl+I'))
        self.italic_button.setToolTip('Italic')
        self.italic_button.clicked.connect(self.set_italic)

        self.underline_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-text-underline.png').getQIcon()
        self.underline_button.setIcon(icon)
        self.underline_button.setAutoRaise(True)
        self.underline_button.setCheckable(True)
        self.underline_button.setFocusPolicy( Qt.ClickFocus )
        self.underline_button.setMaximumSize(QtCore.QSize(20, 20))
        self.underline_button.setShortcut(QtGui.QKeySequence('Ctrl+U'))
        self.underline_button.setToolTip('Underline')
        self.underline_button.clicked.connect(self.set_underline)

        self.copy_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/edit-copy.png').getQIcon()
        self.copy_button.setIcon(icon)
        self.copy_button.setAutoRaise(True)
        self.copy_button.setMaximumSize(QtCore.QSize(20, 20))
        self.copy_button.setFocusPolicy( Qt.ClickFocus )
        self.copy_button.setToolTip('Copy')
        self.copy_button.clicked.connect(self.textedit.copy)

        self.cut_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/edit-cut.png').getQIcon()
        self.cut_button.setIcon(icon)
        self.cut_button.setAutoRaise(True)
        self.cut_button.setMaximumSize(QtCore.QSize(20, 20))
        self.cut_button.setToolTip('Cut')
        self.cut_button.clicked.connect(self.textedit.cut)
        self.cut_button.setFocusPolicy( Qt.ClickFocus )

        self.paste_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/edit-paste.png').getQIcon()
        self.paste_button.setIcon(icon)
        self.paste_button.setAutoRaise(True)
        self.paste_button.setMaximumSize(QtCore.QSize(20, 20))
        self.paste_button.setFocusPolicy( Qt.ClickFocus )
        self.paste_button.setToolTip('Paste')
        self.paste_button.clicked.connect(self.textedit.paste)

        self.alignleft_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-justify-left.png').getQIcon()
        self.alignleft_button.setIcon(icon)
        self.alignleft_button.setAutoRaise(True)
        self.alignleft_button.setCheckable(True)
        self.alignleft_button.setMaximumSize(QtCore.QSize(20, 20))
        self.alignleft_button.setFocusPolicy( Qt.ClickFocus )
        self.alignleft_button.setToolTip('Align left')
        self.alignleft_button.clicked.connect(self.set_alignleft)

        self.aligncenter_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-justify-center.png').getQIcon()
        self.aligncenter_button.setIcon(icon)
        self.aligncenter_button.setAutoRaise(True)
        self.aligncenter_button.setCheckable(True)
        self.aligncenter_button.setMaximumSize(QtCore.QSize(20, 20))
        self.aligncenter_button.setFocusPolicy( Qt.ClickFocus )
        self.aligncenter_button.setToolTip('Align Center')
        self.aligncenter_button.clicked.connect(self.set_aligncenter)

        self.alignright_button = QtGui.QToolButton(self)
        icon = Icon('tango/16x16/actions/format-justify-right.png').getQIcon()
        self.alignright_button.setIcon(icon)
        self.alignright_button.setAutoRaise(True)
        self.alignright_button.setCheckable(True)
        self.alignright_button.setMaximumSize(QtCore.QSize(20, 20))
        self.alignright_button.setFocusPolicy( Qt.ClickFocus )
        self.alignright_button.setToolTip('Align Right')
        self.alignright_button.clicked.connect(self.set_alignright)

        self.color_button = QtGui.QToolButton(self)
        self.color_button.setAutoRaise(True)
        self.color_button.setMaximumSize(QtCore.QSize(20, 20))
        self.color_button.setFocusPolicy( Qt.ClickFocus )
        self.color_button.setToolTip('Color')
        self.color_button.clicked.connect(self.set_color)

        self.toolbar.addWidget(self.copy_button)
        self.toolbar.addWidget(self.cut_button)
        self.toolbar.addWidget(self.paste_button)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.bold_button)
        self.toolbar.addWidget(self.italic_button)
        self.toolbar.addWidget(self.underline_button)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.alignleft_button)
        self.toolbar.addWidget(self.aligncenter_button)
        self.toolbar.addWidget(self.alignright_button)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.color_button)

    #
    # Button methods
    #
    def set_bold(self):
        if self.bold_button.isChecked():
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontWeight(QtGui.QFont.Bold)
        else:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontWeight(QtGui.QFont.Normal)

    def set_italic(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontItalic(True)
        else:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontItalic(False)

    def set_underline(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontUnderline(True)
        else:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setFontUnderline(False)


    def set_alignleft(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setAlignment(Qt.AlignLeft)
        self.update_alignment(Qt.AlignLeft)

    def set_aligncenter(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setAlignment(Qt.AlignCenter)
        self.update_alignment(Qt.AlignCenter)

    def set_alignright(self, bool):
        if bool:
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setAlignment(Qt.AlignRight)
        self.update_alignment(Qt.AlignRight)

    def update_alignment(self, al=None):
        if al is None:
            al = self.textedit.alignment()
        if al == Qt.AlignLeft:
            self.alignleft_button.setChecked(True)
            self.aligncenter_button.setChecked(False)
            self.alignright_button.setChecked(False)
        elif al == Qt.AlignCenter:
            self.aligncenter_button.setChecked(True)
            self.alignleft_button.setChecked(False)
            self.alignright_button.setChecked(False)
        elif al == Qt.AlignRight:
            self.alignright_button.setChecked(True)
            self.alignleft_button.setChecked(False)
            self.aligncenter_button.setChecked(False)

    def set_color(self):
        color = QtGui.QColorDialog.getColor(self.textedit.textColor())
        if color.isValid():
            self.textedit.setFocus(Qt.OtherFocusReason)
            self.textedit.setTextColor(color)
            pixmap = QtGui.QPixmap(16, 16)
            pixmap.fill(color)
            self.color_button.setIcon(QtGui.QIcon(pixmap))

    def update_color(self):
        color = self.textedit.textColor()
        pixmap = QtGui.QPixmap(16, 16)
        pixmap.fill(color)
        self.color_button.setIcon(QtGui.QIcon(pixmap))

    def update_format(self, format):
        font = format.font()
        self.bold_button.setChecked(font.bold())
        self.italic_button.setChecked(font.italic())
        self.underline_button.setChecked(font.underline())
        self.update_alignment(self.textedit.alignment())

    def update_text(self):
        self.update_alignment()
        self.update_color()

    def get_value(self):
        from xml.dom import minidom
        tree = minidom.parseString(unicode(self.textedit.toHtml()).encode('utf-8'))
        value = u''.join([node.toxml() for node in tree.getElementsByTagName('html')[0].getElementsByTagName('body')[0].childNodes])
        return CustomEditor.get_value(self) or value

    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value!=None:
            if unicode(self.textedit.toHtml())!=value:
                self.update_alignment()
                self.textedit.setHtml(value)
                self.update_color()
        else:
            self.textedit.clear()





import tempfile
import unittest
import os

class BinCase(unittest.TestCase):
    """test functions from camelot.bin
    """
            
    def test_create_new_project(self):
        from camelot.bin.meta import CreateNewProject, templates, NewProjectOptions
        new_project_action = CreateNewProject()
        options = NewProjectOptions()
        options.source = tempfile.mkdtemp()
        new_project_action.start_project( options )
        #
        # validate the generated files
        #
        for filename, _template in templates:
            code = open( os.path.join( options.source, 
                                       filename.replace('{{options.module}}', options.module) ) ).read()
            compile( code, 
                     filename,
                     'exec' )
        



from PyQt4 import QtGui

from camelot.view import forms
from camelot.admin.entity_admin import EntityAdmin

class CustomForm(forms.Form):
    
    def __init__(self):
        super(CustomForm, self).__init__(['first_name', 'last_name'])
        
    def render( self, widgets, parent = None, nomargins = False ):
        widget = QtGui.QWidget(parent)
        layout = QtGui.QFormLayout()
        layout.addRow(QtGui.QLabel('Please fill in the complete name :', widget))
        for _field_name,(field_label, field_editor) in widgets.items():
            layout.addRow(field_label, field_editor)
        widget.setLayout(layout)
        widget.setBackgroundRole(QtGui.QPalette.ToolTipBase)
        widget.setAutoFillBackground(True)
        return widget

class Admin(EntityAdmin):
    list_display = ['first_name', 'last_name']
    form_display = CustomForm()
    form_size = (300,100)


#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

'''
Created on May 22, 2010

@author: tw55413
'''

from PyQt4 import QtGui
from PyQt4 import QtCore

from camelot.admin.action.form_action import FormActionGuiContext
from camelot.view.model_thread import post

class AbstractActionWidget( object ):

    def __init__( self, action, gui_context ):
        """Helper class to construct widget that when triggered run an action.
        This class exists as a base class for custom ActionButton implementations.
        """
        from camelot.admin.action import State
        self.action = action
        self.gui_context = gui_context
        self.state = State()
        if isinstance( gui_context, FormActionGuiContext ):
            gui_context.widget_mapper.model().dataChanged.connect( self.data_changed )
            gui_context.widget_mapper.currentIndexChanged.connect( self.current_row_changed )
        post( action.get_state, self.set_state, args = (self.gui_context.create_model_context(),) )

    def set_state( self, state ):
        self.state = state
        self.setEnabled( state.enabled )
        self.setVisible( state.visible )
        
    def current_row_changed( self, current_row ):
        post( self.action.get_state, 
              self.set_state, 
              args = (self.gui_context.create_model_context(),) )
        
    def data_changed( self, index1, index2 ):
        self.current_row_changed( index1.row() )
        
    def run_action( self, mode=None ):
        gui_context = self.gui_context.copy()
        gui_context.mode = mode
        self.action.gui_run( gui_context )

HOVER_ANIMATION_DISTANCE = 20
NOTIFICATION_ANIMATION_DISTANCE = 8

class ActionLabel( QtGui.QLabel, AbstractActionWidget ):
    
    entered = QtCore.pyqtSignal()
    left = QtCore.pyqtSignal()    
    
    """
    A custom interactive desktop button for the desktop. Each 'button' is 
    actually an animated label.
    """
    def __init__( self, action, gui_context, parent ):
        QtGui.QLabel.__init__( self, parent )
        AbstractActionWidget.__init__( self, action, gui_context )
        
        self.setObjectName('ActionButton')
        self.setMouseTracking(True)
        
        # This property holds if this button reacts to mouse events.
        self.interactive = False
        
        # This property is used to store the original position of this label
        # so it can be visually reset when the user leaves before the ongoing
        # animation has finished.
        self.originalPosition = None

        # This property holds the state of the selection animation. Since this
        # animation is only created inside startSelectionAnimation() (to avoid
        # the increasing amount of performAction() invocations), this variable is 
        # used to continuously store the state of that animation.
        self.selectionAnimationState = QtCore.QAbstractAnimation.Stopped
        self.setMaximumHeight(160)
        
        opacityEffect = QtGui.QGraphicsOpacityEffect(parent = self)
        opacityEffect.setOpacity(1.0)
        self.setGraphicsEffect(opacityEffect)
    
        # Bounce animation #
        hoverAnimationPart1 = QtCore.QPropertyAnimation(self, 'pos')
        hoverAnimationPart1.setObjectName('hoverAnimationPart1')
        hoverAnimationPart1.setDuration(500)
        hoverAnimationPart1.setEasingCurve(QtCore.QEasingCurve.Linear)
        
        hoverAnimationPart2 = QtCore.QPropertyAnimation(self, 'pos')
        hoverAnimationPart2.setObjectName('hoverAnimationPart2')
        hoverAnimationPart2.setDuration(1500)
        hoverAnimationPart2.setEasingCurve(QtCore.QEasingCurve.OutElastic)
        
        hoverAnimation = QtCore.QSequentialAnimationGroup(parent = self)
        hoverAnimation.setObjectName('hoverAnimation')
        hoverAnimation.setLoopCount(-1) # Infinite
        hoverAnimation.addAnimation(hoverAnimationPart1)
        hoverAnimation.addAnimation(hoverAnimationPart2)
        ####################
        
        # Selection animation #
        selectionAnimationPart1 = QtCore.QPropertyAnimation(self, 'pos')
        selectionAnimationPart1.setObjectName('selectionAnimationPart1')
        selectionAnimationPart1.setDuration(200)
        selectionAnimationPart1.setEasingCurve(QtCore.QEasingCurve.Linear)
        
        selectionAnimationPart2 = QtCore.QPropertyAnimation(self, 'size')
        selectionAnimationPart2.setObjectName('selectionAnimationPart2')
        selectionAnimationPart2.setDuration(200)
        selectionAnimationPart2.setEasingCurve(QtCore.QEasingCurve.OutCubic)
        
        selectionAnimationPart3 = QtCore.QPropertyAnimation(self.graphicsEffect(), 'opacity')
        selectionAnimationPart3.setObjectName('selectionAnimationPart3')
        selectionAnimationPart3.setDuration(200)
        selectionAnimationPart3.setEasingCurve(QtCore.QEasingCurve.Linear)
        
        selectionAnimation = QtCore.QParallelAnimationGroup(parent = self)
        selectionAnimation.setObjectName('selectionAnimation')
        selectionAnimation.addAnimation(selectionAnimationPart1)
        selectionAnimation.addAnimation(selectionAnimationPart2)
        selectionAnimation.addAnimation(selectionAnimationPart3)
        # Not working when clicking the white area underneath the ActionButton image.
        #selectionAnimation.finished.connect(self.resetLayout)
        #selectionAnimation.finished.connect(self.performAction)
        selectionAnimation.stateChanged.connect(self.updateSelectionAnimationState)
        #######################

    def set_state( self, state ):
        AbstractActionWidget.set_state( self, state )
        if state.icon:
            self.setPixmap( state.icon.getQPixmap() )
            self.resize( self.pixmap().width(), self.pixmap().height() )
        if state.notification:
            # Shake animation #
            notificationAnimationPart1 = QtCore.QPropertyAnimation(self, 'pos')
            notificationAnimationPart1.setObjectName('notificationAnimationPart1')
            notificationAnimationPart1.setDuration(50)
            notificationAnimationPart1.setEasingCurve(QtCore.QEasingCurve.Linear)
            
            notificationAnimationPart2 = QtCore.QPropertyAnimation(self, 'pos')
            notificationAnimationPart2.setObjectName('notificationAnimationPart2')
            notificationAnimationPart2.setDuration(50)
            notificationAnimationPart2.setEasingCurve(QtCore.QEasingCurve.Linear)
            
            notificationAnimationPart3 = QtCore.QPropertyAnimation(self, 'pos')
            notificationAnimationPart3.setObjectName('notificationAnimationPart3')
            notificationAnimationPart3.setDuration(50)
            notificationAnimationPart3.setEasingCurve(QtCore.QEasingCurve.Linear)
            
            notificationAnimation = QtCore.QSequentialAnimationGroup(parent = self)
            notificationAnimation.setObjectName('notificationAnimation')
            notificationAnimation.setLoopCount(10)
            notificationAnimation.addAnimation(notificationAnimationPart1)
            notificationAnimation.addAnimation(notificationAnimationPart2)
            notificationAnimation.addAnimation(notificationAnimationPart3)

            # Timer is used to simulate a pausing effect of the animation.
            notificationAnimationTimer = QtCore.QTimer(parent = self)
            notificationAnimationTimer.setObjectName('notificationAnimationTimer')
            notificationAnimationTimer.setInterval(1500)
            notificationAnimationTimer.setSingleShot(True)
            notificationAnimationTimer.timeout.connect(notificationAnimation.start)
            notificationAnimation.finished.connect( notificationAnimationTimer.start )
        self.resetLayout()
        
    def startHoverAnimation(self):
        hoverAnimationPart1 = self.findChild(QtCore.QPropertyAnimation, 'hoverAnimationPart1')
        if hoverAnimationPart1 is not None:
            hoverAnimationPart1.setStartValue(self.originalPosition)
            hoverAnimationPart1.setEndValue(self.originalPosition + QtCore.QPoint(0, -HOVER_ANIMATION_DISTANCE))
    
        hoverAnimationPart2 = self.findChild(QtCore.QPropertyAnimation, 'hoverAnimationPart2')
        if hoverAnimationPart2 is not None:
            hoverAnimationPart2.setStartValue(self.originalPosition + QtCore.QPoint(0, -HOVER_ANIMATION_DISTANCE))
            hoverAnimationPart2.setEndValue(self.originalPosition)
        
        hoverAnimation = self.findChild(QtCore.QSequentialAnimationGroup, 'hoverAnimation')
        if hoverAnimation is not None:
            hoverAnimation.start()

    def stopHoverAnimation(self):
        hoverAnimation = self.findChild(QtCore.QSequentialAnimationGroup, 'hoverAnimation')
        if hoverAnimation is not None:
            hoverAnimation.stop()
        if self.originalPosition is not None:
            self.move(self.originalPosition)
            
        self.resetLayout()

    def startSelectionAnimation(self):
        if self.state.notification:
            notificationAnimation = self.findChild(QtCore.QSequentialAnimationGroup, 'notificationAnimation')            
            if notificationAnimation is not None:
                notificationAnimation.stop()
        else:
            hoverAnimation = self.findChild(QtCore.QSequentialAnimationGroup, 'hoverAnimation')
            if hoverAnimation is not None:
                hoverAnimation.stop()

        self.move(self.originalPosition)
        
        # Selection animation when clicked #
        selectionAnimationPart1 = self.findChild(QtCore.QPropertyAnimation, 'selectionAnimationPart1')
        selectionAnimationPart2 = self.findChild(QtCore.QPropertyAnimation, 'selectionAnimationPart2')
        selectionAnimationPart3 = self.findChild(QtCore.QPropertyAnimation, 'selectionAnimationPart3')
        selectionAnimation = self.findChild(QtCore.QParallelAnimationGroup, 'selectionAnimation')
        if None not in (selectionAnimationPart1, selectionAnimationPart2, 
                        selectionAnimationPart3, selectionAnimation):
            selectionAnimationPart1.setStartValue(self.originalPosition)
            selectionAnimationPart1.setEndValue(self.originalPosition + QtCore.QPoint(-HOVER_ANIMATION_DISTANCE, -HOVER_ANIMATION_DISTANCE))

            selectionAnimationPart2.setStartValue(self.size())
            selectionAnimationPart2.setEndValue(self.size() + QtCore.QSize(40, 40))

            selectionAnimationPart3.setStartValue(1.0)
            selectionAnimationPart3.setEndValue(0.1)
        
            self.setScaledContents(True)
            
            selectionAnimation.start()

    def updateSelectionAnimationState(self, newState, oldState):
        self.selectionAnimationState = newState
        
        # Simulate finished signal (see comment in animation buildup code).
        if oldState == QtCore.QAbstractAnimation.Running and newState == QtCore.QAbstractAnimation.Stopped:
            self.run_action()
            self.resetLayout()

    def resetLayout(self):
        if self.state.notification:
            self.stopNotificationAnimation()
        
        if self.sender() and self.originalPosition:
            self.move(self.originalPosition)

        self.setScaledContents(False)
        if self.pixmap():
            self.resize(self.pixmap().width(), self.pixmap().height())
        self.graphicsEffect().setOpacity(1.0)
        
        if self.state.notification and self.originalPosition:
            self.startNotificationAnimation()
        
    def setInteractive(self, interactive):
        self.interactive = interactive
        
        self.originalPosition = self.mapToParent(QtCore.QPoint(0, 0))# + QtCore.QPoint(NOTIFICATION_ANIMATION_DISTANCE, HOVER_ANIMATION_DISTANCE)
        
        if self.state.notification:
            self.startNotificationAnimation()

    def enterEvent(self, event):
        if self.interactive:
            if self.state.notification:
                self.stopNotificationAnimation()
                
            self.startHoverAnimation()
            
            self.entered.emit()

        event.ignore()
    
    def leaveEvent(self, event):
        if self.interactive:
            self.stopHoverAnimation()
            
            if self.state.notification:
                self.startNotificationAnimation()
            
            self.left.emit()

        event.ignore()

    def onContainerMousePressEvent(self, event):
        # Second part blocks fast consecutive clicks.
        if self.interactive and self.selectionAnimationState == QtCore.QAbstractAnimation.Stopped:
            self.startSelectionAnimation()

        event.ignore()
        
    def startNotificationAnimation(self):
        notificationAnimationPart1 = self.findChild(QtCore.QPropertyAnimation, 'notificationAnimationPart1')
        if notificationAnimationPart1 is not None:
            notificationAnimationPart1.setStartValue(self.originalPosition)
            notificationAnimationPart1.setEndValue(self.originalPosition + QtCore.QPoint(-NOTIFICATION_ANIMATION_DISTANCE, 0))

        notificationAnimationPart2 = self.findChild(QtCore.QPropertyAnimation, 'notificationAnimationPart2')
        if notificationAnimationPart2 is not None:
            notificationAnimationPart2.setStartValue(self.originalPosition + QtCore.QPoint(-NOTIFICATION_ANIMATION_DISTANCE, 0))
            notificationAnimationPart2.setEndValue(self.originalPosition + QtCore.QPoint(NOTIFICATION_ANIMATION_DISTANCE, 0))
        
        notificationAnimationPart3 = self.findChild(QtCore.QPropertyAnimation, 'notificationAnimationPart3')
        if notificationAnimationPart3 is not None:
            notificationAnimationPart3.setStartValue(self.originalPosition + QtCore.QPoint(NOTIFICATION_ANIMATION_DISTANCE, 0))
            notificationAnimationPart3.setEndValue(self.originalPosition)
        
        notificationAnimation = self.findChild(QtCore.QSequentialAnimationGroup, 'notificationAnimation')
        notificationAnimationTimer = self.findChild(QtCore.QTimer, 'notificationAnimationTimer')
        if notificationAnimation is not None and notificationAnimationTimer is not None:
            notificationAnimationTimer.start()
            notificationAnimation.start()

    def stopNotificationAnimation(self):
        notificationAnimation = self.findChild(QtCore.QSequentialAnimationGroup, 'notificationAnimation')
        notificationAnimationTimer = self.findChild(QtCore.QTimer, 'notificationAnimationTimer')
        
        if notificationAnimation is not None and notificationAnimationTimer is not None:
            notificationAnimationTimer.stop()
            notificationAnimation.stop()

class ActionPushButton( QtGui.QPushButton, AbstractActionWidget ):
    
    def __init__( self, action, gui_context, parent ):
        """A :class:`QtGui.QPushButton` that when pressed, will run an 
        action.
        
        .. image:: /_static/actionwidgets/action_push_botton_application_enabled.png
        
        """
        QtGui.QPushButton.__init__( self,
                                    parent )
        AbstractActionWidget.__init__( self, action, gui_context )
        self.clicked.connect( self.triggered )

    @QtCore.pyqtSlot()
    def triggered(self):
        self.run_action( None )
        
    @QtCore.pyqtSlot( QtCore.QModelIndex, QtCore.QModelIndex )
    def data_changed( self, index1, index2 ):
        AbstractActionWidget.data_changed( self, index1, index2 )
        
    def set_state( self, state ):
        super( ActionPushButton, self ).set_state( state )
        if state.verbose_name != None:
            self.setText( unicode( state.verbose_name ) )
        if state.icon != None:
            self.setIcon( state.icon.getQIcon() )
        else:
            self.setIcon( QtGui.QIcon() )
        if state.modes:
            menu = QtGui.QMenu( self )
            for mode in state.modes:
                menu.addAction( mode.render( menu ) )
            self.setMenu( menu )



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from application_action import ( ApplicationActionGuiContext,
                                 OpenNewView, OpenTableView)
from list_action import ListActionGuiContext, CallMethod, OpenFormView

from base import Action, ActionStep, GuiContext, Mode, State

__all__ = [
    Action.__name__,
    ActionStep.__name__,
    ApplicationActionGuiContext.__name__,
    CallMethod.__name__,
    ListActionGuiContext.__init__,
    OpenFormView.__init__,
    OpenNewView.__name__,
    OpenTableView.__name__,
    GuiContext.__name__,
    Mode.__name__,
    State.__name__,
    ]



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from camelot.view.model_thread import post
from choiceseditor import ChoicesEditor

class OneToManyChoicesEditor(ChoicesEditor):
  
    def __init__(self, 
                 parent, 
                 target=None, 
                 nullable=True, 
                 field_name='onetomanychoices',
                 **kwargs):
        super(OneToManyChoicesEditor, self).__init__(parent, **kwargs)
        self.setObjectName( field_name )
        assert target!=None
        self._target = target
        self._nullable = nullable
        post(self.get_choices, self.set_choices)

    def get_choices(self):
        additional_choices = []
        if self._nullable:
            additional_choices = [(None, '')]
        return additional_choices + [(o, unicode(o)) for o in self._target.query.all()]
        
    def set_field_attributes(self, editable=True, **kwargs):
        """Makes sure choices are not reset when changing the
        field attributes"""
        self.setEnabled(editable!=False)




#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import base64
import logging

from PyQt4 import QtCore

from camelot.core.conf import settings

logger = logging.getLogger('camelot.core.dbprofiles')

def get_cipher():
    from Crypto.Cipher import ARC4
    if hasattr( settings, 'CAMELOT_DBPROFILES_CIPHER' ):
        key = getattr( settings, 'CAMELOT_DBPROFILES_CIPHER' )
    else:
        key = 'The Knights Who Say Ni'
    return ARC4.new( key )

def get_languagecode(profile=None):
    """
    :return: two-letter ISO 639 language code
    """
    if not profile:
        profile = selected_profile_info()
    return selected_profile_info()['locale_language'][:2]

def get_countrycode(profile=None):
    """
    :return: two-letter ISO 3166 country code
    """
    if not profile:
        profile = selected_profile_info()
    return selected_profile_info()['locale_language'][2:]

def _encode_setting(value):
    return base64.b64encode( get_cipher().encrypt( unicode(value).encode('utf-8' ) ) )

def _decode_setting(value):
    return get_cipher().decrypt( base64.b64decode( value ) ).decode('utf-8')

def selected_profile_info():
    """
    :return: a dict with the info of the selected profile
    """
    profiles = fetch_profiles()
    profilename = last_used_profile()
    try:
        return profiles[profilename]
    except KeyError:
        logger.error( u'no profile named %s, available profiles are '%profilename )
        for key in profiles.keys():
            logger.error( u' - %s'%key )
        raise

def engine_from_profile():
    from sqlalchemy import create_engine
    profile = selected_profile_info()
    connect_args = dict()
    if profile['dialect'] == 'mysql':
        connect_args['charset'] = 'utf8'

    connection = '%s://%s:%s@%s:%s/%s' % (profile['dialect'],
                                          profile['user'],
                                          profile['pass'],
                                          profile['host'],
                                          profile['port'],
                                          profile['database'])
    return create_engine(connection, pool_recycle=True, connect_args=connect_args)

def media_root_from_profile():
    profile = selected_profile_info()
    return profile['media_location']

def stylesheet_from_profile():
    profile = selected_profile_info()
    from camelot.view import art
    return art.read( 'stylesheet/office2007_' + profile.get('stylesheet', 'blue') + '.qss' )

def last_used_profile():
    settings = QtCore.QSettings()
    return unicode(settings.value('last_used_database_profile',
        QtCore.QVariant('')).toString(), 'utf-8')

def fetch_profiles(from_file=None):
    profiles = {}
    try:
        if from_file is None:
            settings = QtCore.QSettings()
        else:
            settings = QtCore.QSettings(from_file, QtCore.QSettings.IniFormat)

        size = settings.beginReadArray('database_profiles')

        if size == 0:
            return profiles

        for index in range(size):
            settings.setArrayIndex(index)
            info = {}
            profilename = unicode(settings.value('profilename', QtCore.QVariant('')).toString(), 'utf-8')
            if not profilename:
                continue  # well we should not really be doing anything
            info['dialect'] = _decode_setting(settings.value('dialect', QtCore.QVariant('')).toString())
            info['host'] = _decode_setting(settings.value('host', QtCore.QVariant('')).toString())
            info['port'] = _decode_setting(settings.value('port', QtCore.QVariant('')).toString())
            info['database'] = _decode_setting(settings.value('database', QtCore.QVariant('')).toString())
            info['user'] = _decode_setting(settings.value('user', QtCore.QVariant('')).toString())
            info['pass'] = _decode_setting(settings.value('pass', QtCore.QVariant('')).toString())
            info['media_location'] = _decode_setting(settings.value('media_location', QtCore.QVariant('')).toString())
            info['locale_language'] = _decode_setting(settings.value('locale_language', QtCore.QVariant('')).toString())
            info['proxy_host'] = _decode_setting(settings.value('proxy_host', QtCore.QVariant('')).toString())
            info['proxy_port'] = _decode_setting(settings.value('proxy_port', QtCore.QVariant('')).toString())
            info['proxy_username'] = _decode_setting(settings.value('proxy_username', QtCore.QVariant('')).toString())
            info['proxy_password'] = _decode_setting(settings.value('proxy_password', QtCore.QVariant('')).toString())
            profiles[profilename] = info
        settings.endArray()
    except Exception, e:
        logger.warn('Could not read existing profiles, proceed with what was available', exc_info=e)
    return profiles

def store_profiles(profiles, to_file=None):
    if to_file is None:
        settings = QtCore.QSettings()
    else:
        settings = QtCore.QSettings(to_file, QtCore.QSettings.IniFormat)

    settings.beginWriteArray('database_profiles')

    for index, (profilename, info) in enumerate(profiles.items()):
        settings.setArrayIndex(index)
        settings.setValue('profilename', QtCore.QVariant(unicode(profilename).encode('utf-8')))
        settings.setValue('dialect', QtCore.QVariant(_encode_setting(info['dialect'])))
        settings.setValue('host', QtCore.QVariant(_encode_setting(info['host'])))
        settings.setValue('port', QtCore.QVariant(_encode_setting(info['port'])))
        settings.setValue('database', QtCore.QVariant(_encode_setting(info['database'])))
        settings.setValue('user', QtCore.QVariant(_encode_setting(info['user'])))
        settings.setValue('pass', QtCore.QVariant(_encode_setting(info['pass'])))
        settings.setValue('media_location', QtCore.QVariant(_encode_setting(info['media_location'])))
        settings.setValue('locale_language', QtCore.QVariant(_encode_setting(info['locale_language'])))
        settings.setValue('proxy_host', QtCore.QVariant(_encode_setting(info['proxy_host'])))
        settings.setValue('proxy_port', QtCore.QVariant(_encode_setting(info['proxy_port'])))
        settings.setValue('proxy_username', QtCore.QVariant(_encode_setting(info['proxy_username'])))
        settings.setValue('proxy_password', QtCore.QVariant(_encode_setting(info['proxy_password'])))
    settings.endArray()

def use_chosen_profile(profilename):
    settings = QtCore.QSettings()
    settings.setValue('last_used_database_profile', unicode(profilename).encode('utf-8') )

class EmptyProxy():

    @classmethod
    def hostName(cls):
        return ''

    @classmethod
    def user(cls):
        return ''

    @classmethod
    def port(cls):
        return ''

    @classmethod
    def password(cls):
        return ''

def get_network_proxy():
    # turn this temporary off, because it freezes the app on winblows
    return EmptyProxy()

    #from PyQt4 import QtNetwork

    #proxy = None
    #query = QtNetwork.QNetworkProxyQuery(QtCore.QUrl('http://aws.amazon.com'))
    ##proxies = QtNetwork.QNetworkProxyFactory.systemProxyForQuery(query)

    #if proxies:
        #logger.info('Proxy servers found: %s' % ['%s:%s' %
            #(str(proxy.hostName()),str(proxy.port())) for proxy in proxies])
        #if proxies[0].hostName():
            #proxy = proxies[0]

    ## we still need some empty values for the profile
    #if proxy is None:
        #return EmptyProxy()

    #return proxy



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import datetime

from PyQt4 import QtGui
from PyQt4 import QtCore
from PyQt4.QtCore import Qt

from customeditor import CustomEditor, set_background_color_palette

from camelot.view.art import Icon
from camelot.view.utils import local_date_format, date_from_string, ParsingError
from camelot.view.controls.decorated_line_edit import DecoratedLineEdit
from camelot.core.utils import ugettext as _

class DateEditor(CustomEditor):
    """Widget for editing date values"""

    calendar_action_trigger = QtCore.pyqtSignal()
    special_date_icon = Icon('tango/16x16/apps/office-calendar.png')
    
    def __init__(self, parent = None,
                       editable = True,
                       nullable = True, 
                       field_name = 'date',
                       **kwargs):
        CustomEditor.__init__(self, parent)

        self.setObjectName( field_name )
        self.date_format = local_date_format()
        self.line_edit = DecoratedLineEdit()
        self.line_edit.set_minimum_width( len( self.date_format ) )
        self.line_edit.set_background_text( QtCore.QDate(2000,1,1).toString(self.date_format) )

        # The order of creation of this widgets and their parenting
        # seems very sensitive under windows and creates system crashes
        # so don't change this without extensive testing on windows
        special_date_menu = QtGui.QMenu(self)
        calendar_widget_action = QtGui.QWidgetAction(special_date_menu)
        self.calendar_widget = QtGui.QCalendarWidget(special_date_menu)
        self.calendar_widget.activated.connect(self.calendar_widget_activated)
        self.calendar_widget.clicked.connect(self.calendar_widget_activated)
        calendar_widget_action.setDefaultWidget(self.calendar_widget)

        self.calendar_action_trigger.connect( special_date_menu.hide )
        special_date_menu.addAction(calendar_widget_action)
        special_date_menu.addAction(_('Today'))
        special_date_menu.addAction(_('Far future'))
        self.special_date = QtGui.QToolButton(self)
        self.special_date.setIcon( self.special_date_icon.getQIcon() )
        self.special_date.setAutoRaise(True)
        self.special_date.setToolTip(_('Calendar and special dates'))
        self.special_date.setMenu(special_date_menu)
        self.special_date.setPopupMode(QtGui.QToolButton.InstantPopup)
        self.special_date.setFixedHeight(self.get_height())
        self.special_date.setFocusPolicy(Qt.ClickFocus)
        # end of sensitive part

        if nullable:
            special_date_menu.addAction(_('Clear'))

        self.hlayout = QtGui.QHBoxLayout()
        self.hlayout.addWidget(self.line_edit)
        self.hlayout.addWidget(self.special_date)

        self.hlayout.setContentsMargins(0, 0, 0, 0)
        self.hlayout.setSpacing(0)
        self.hlayout.setAlignment(Qt.AlignRight|Qt.AlignVCenter)
        
        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.hlayout)

        self.minimum = datetime.date.min
        self.maximum = datetime.date.max
        self.setFocusProxy(self.line_edit)

        self.line_edit.editingFinished.connect( self.line_edit_finished )
        self.line_edit.textEdited.connect(self.text_edited)
        special_date_menu.triggered.connect(self.set_special_date)

    def calendar_widget_activated(self, date):
        self.calendar_action_trigger.emit()
        self.set_value(date)
        self.editingFinished.emit()
        self.line_edit.setFocus()

    def line_edit_finished(self):
        self.setProperty( 'value', QtCore.QVariant( self.get_value() ) )
        self.valueChanged.emit()
        self.editingFinished.emit()

    def focusOutEvent(self, event):
        # explicitely set value on focus out to format the date in case
        # it was entered unformatted
        value = self.get_value()
        self.set_value( value )
        self.editingFinished.emit()

    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        self.setProperty( 'value', QtCore.QVariant( value ) )
        if value:
            qdate = QtCore.QDate(value)
            formatted_date = qdate.toString(self.date_format)
            self.line_edit.set_user_input(formatted_date)
            self.calendar_widget.setSelectedDate(qdate)
        else:
            self.line_edit.set_user_input('')
        self.valueChanged.emit()

    def text_edited(self, text ):
        try:
            date_from_string( self.line_edit.user_input() )
            self.line_edit.set_valid(True)
            self.valueChanged.emit()
        except ParsingError:
            self.line_edit.set_valid(False)

    def get_value(self):
        try:
            value = date_from_string( self.line_edit.user_input() )
        except ParsingError:
            value = None
        return CustomEditor.get_value(self) or value

    def set_field_attributes(self, editable = True,
                                   background_color = None,
                                   tooltip = None, **kwargs):
        self.set_enabled(editable)
        self.set_background_color(background_color)
        self.line_edit.setToolTip(unicode(tooltip or ''))

    def set_background_color(self, background_color):
        set_background_color_palette( self.line_edit, background_color )

    def set_enabled(self, editable=True):
        self.line_edit.setEnabled(editable)
        if editable:
            self.special_date.show()
        else:
            self.special_date.hide()

    def set_special_date(self, action):
        if action.text().compare(_('Today')) == 0:
            self.set_value(datetime.date.today())
        elif action.text().compare(_('Far future')) == 0:
            self.set_value(datetime.date( year = 2400, month = 12, day = 31 ))
        elif action.text().compare(_('Clear')) == 0:
            self.set_value(None)
        self.line_edit.setFocus()
        self.editingFinished.emit()



from elixir import Entity, Field, ManyToOne
from sqlalchemy.types import Unicode, Date
from camelot.admin.entity_admin import EntityAdmin
from camelot.view import forms

class Movie(Entity):
    title = Field(Unicode(60), required=True)
    short_description = Field(Unicode(512))
    releasedate = Field(Date)
    director = ManyToOne('Person')
  
    class Admin(EntityAdmin):
        form_display = forms.Form(['title', 'short_description', 'director', 'releasedate'])



from camelot.admin.entity_admin import EntityAdmin
from camelot.core.utils import ugettext_lazy as _

class VisitorsPerDirector(object):
    
    class Admin(EntityAdmin):
        verbose_name = _('Visitors per director')
        list_display = ['first_name', 'last_name', 'visitors']
        

def setup_views():
    from sqlalchemy.sql import select, func, and_
    from sqlalchemy.orm import mapper
 
    from camelot.model.authentication import Person
    from camelot_example.model import Movie, VisitorReport
    
    s = select([Person.party_id,
                Person.first_name.label('first_name'),
                Person.last_name.label('last_name'),
                func.sum( VisitorReport.visitors ).label('visitors'),],
                whereclause = and_( Person.party_id == Movie.director_party_id,
                                    Movie.id == VisitorReport.movie_id),
                group_by = [ Person.party_id, 
                             Person.first_name, 
                             Person.last_name ] )
                            
    s=s.alias('visitors_per_director')
    
    mapper( VisitorsPerDirector, s, always_refresh=True )



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import os
import logging

from camelot.view.model_thread import post
from camelot.core.utils import ugettext as _
from camelot.view.controls.exception import model_thread_exception_message_box

from PyQt4 import QtGui
from PyQt4 import QtCore

LOGGER = logging.getLogger('camelot.view.storage')

class OpenFileProgressDialog(QtGui.QProgressDialog):

    def __init__(self):
        QtGui.QProgressDialog.__init__(
            self, _('Please wait'), QtCore.QString(), 0, 0
        )
        self.setWindowTitle(_('Open file'))
        self.setRange(0, 0)

    def open_path(self, path):
        if not os.path.exists(path):
            QtGui.QMessageBox.critical(
                self,
                _('Could not open file'),
                _('%s does not exist') % path,
            )
        #
        # support for windows shares
        #
        if not path.startswith(r'\\'):
            url = QtCore.QUrl.fromLocalFile(path)
        else:
            url = QtCore.QUrl(path, QtCore.QUrl.TolerantMode)
        QtGui.QDesktopServices.openUrl(url)
        self.close()

    def file_stored(self):
        """Called when the file has been stored at path"""
        self.close()


class SaveFileProgressDialog(QtGui.QProgressDialog):

    def __init__(self):
        QtGui.QProgressDialog.__init__(
            self, _('Please wait'), QtCore.QString(), 0, 0
        )
        self.setWindowTitle(_('Save file'))
        self.setRange(0, 0)

    def finish(self, on_finish):
        on_finish()
        self.close()


def open_stored_file(parent, stored_file):
    """Open the stored file with the default system editor for
    this file type"""

    progress = OpenFileProgressDialog()

    def get_path():
        return stored_file.storage.checkout(stored_file)

    post(get_path, progress.open_path, model_thread_exception_message_box)
    progress.exec_()

def save_stored_file(parent, stored_file):
    """Save a stored file as another file"""
    settings = QtCore.QSettings()
    default_dir = settings.value('datasource').toString()
    proposal = os.path.join(unicode(default_dir), unicode(stored_file.verbose_name) )

    path = QtGui.QFileDialog.getSaveFileName(
        parent, _('Save as'), proposal
    )

    if path:
        progress = OpenFileProgressDialog()

        def save_as():
            destination = open(path, 'wb')
            destination.write( stored_file.storage.checkout_stream(stored_file).read() )

        post(save_as, progress.file_stored, model_thread_exception_message_box)
        progress.exec_()

def create_stored_file(parent, storage, on_finish, filter='All files (*)',
                       remove_original = False,
                       filename = None):
    """Popup a QFileDialog, put the selected file in the storage and
    return the call on_finish with the StoredFile when done
    
    :param on_finish: function that will be called in the gui thread when
    the file is stored.  the first argument of the function should be the
    StoredFile
    
    :param filename: if None, a dialog will pop up, asking the user for
    the file, otherwise a string with the name of the file to be stored
    """
    settings = QtCore.QSettings()
    dir = settings.value('lastpath').toString()
    # use last path saved in settings, if none current dir is used by Qt
    if filename == None:
        filename = QtGui.QFileDialog.getOpenFileName(
            parent, 'Open file', dir, filter
        )
    if filename:
        filename = unicode( filename )
        remove = False
        if remove_original:
            reply = QtGui.QMessageBox(
                QtGui.QMessageBox.Warning,
                _('The file will be stored.'),
                _('Do you want to remove the original file?'),
                QtGui.QMessageBox.No | QtGui.QMessageBox.Yes,
            )
            reply.setDefaultButton(QtGui.QMessageBox.Yes)
            reply.exec_()
            if reply == QtGui.QMessageBox.Yes:
                remove = True
        # save it back
        settings.setValue('lastpath', QtCore.QVariant( os.path.dirname( filename ) ) )
        progress = SaveFileProgressDialog()

        def checkin():
            new_path = storage.checkin( filename )
            if remove:
                try:
                    os.remove( filename )
                except Exception, e:
                    LOGGER.warn('could not remove file', exc_info=e)
            return lambda:on_finish(new_path)

        post(checkin, progress.finish, model_thread_exception_message_box)
        progress.exec_()



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from PyQt4.QtCore import Qt

from camelot.view.controls.editors import MonthsEditor
from camelot.view.controls.delegates.customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.core.utils import variant_to_pyobject, ugettext
from camelot.view.proxy import ValueLoading

class MonthsDelegate(CustomDelegate):
    """MonthsDelegate

    custom delegate for showing and editing months and years
    """

    editor = MonthsEditor

    __metaclass__ = DocumentationMetaclass

    def __init__(self, parent=None, forever=200*12, **kwargs):
        """
        :param forever: number of months that will be indicated as Forever, set
        to None if not appliceable
        """
        super(MonthsDelegate, self).__init__(parent=parent, **kwargs)
        self._forever = forever
        
    def sizeHint(self, option, index):
        q = MonthsEditor(None)
        return q.sizeHint()

    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        value = variant_to_pyobject( index.model().data( index, Qt.EditRole ) )
        
        value_str = u''
        if self._forever != None and value == self._forever:
            value_str = ugettext('Forever')
        elif value not in (None, ValueLoading):
            years, months = divmod( value, 12 )
            if years:
                value_str = value_str + ugettext('%i years ')%(years)
            if months:
                value_str = value_str + ugettext('%i months')%(months)        

        self.paint_text(painter, option, index, value_str)
        painter.restore()





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from camelot.view.controls import editors
from one2manydelegate import One2ManyDelegate

class ManyToManyDelegate(One2ManyDelegate):
    """
  .. image:: /_static/manytomany.png
  """

    def createEditor(self, parent, option, index):
        editor = editors.ManyToManyEditor(parent=parent, **self.kwargs)
        self.setEditorData(editor, index)
        editor.editingFinished.connect( self.commitAndCloseEditor )
        return editor

    #@QtCore.pyqtSlot()
    # not yet converted to new style sig slot because sender doesn't work
    # in certain versions of pyqt
    def commitAndCloseEditor(self):
        editor = self.sender()
        self.commitData.emit(editor)

    def setModelData(self, editor, model, index):
        if editor.getModel():
            model.setData(index, editor.getModel().get_collection)






from camelot.test import ModelThreadTestCase

class QueryProxyCase( ModelThreadTestCase ):
  """Test the functionality of the QueryProxy to perform CRUD operations on 
  stand alone data"""
  
  def setUp(self):
    super( QueryProxyCase, self ).setUp()
    from camelot.model.authentication import Person
    from camelot.view.proxy.queryproxy import QueryTableProxy
    from camelot.admin.application_admin import ApplicationAdmin
    self.app_admin = ApplicationAdmin()
    person_admin = self.app_admin.get_related_admin( Person )
    self.person_proxy = QueryTableProxy( person_admin, 
                                         query_getter = None, 
                                         columns_getter = person_admin.get_fields )
  
  def test_sort( self ):
    from camelot.model.authentication import Person
